"""
Functions for making splits
"""

import collections
from abc import ABCMeta

import numpy as np
#from sklearn.externals.six import with_metaclass
from sklearn.model_selection._split import KFold as _KFold
from sklearn.model_selection._split import ShuffleSplit as _ShuffleSplit
from sklearn.utils import check_random_state


def exponential_split(xmin, xmax, n=5):
    """
    Obtain integers that are equally spaced in log space.

    Parameters
    ----------
    xmin: float
          lower bound in original space
    xmax: float
          upper bound in original space
    n: int
       integer giving the number of spaces (default is 5)

    Returns:
    -------
    X: np.array
       list of n evenly spaced points in log space
    """

    X = np.zeros(n, dtype=int)
    [lmin, lmax] = [np.log(xmin), np.log(xmax)]
    dl = (lmax - lmin) / (n - 1.)
    X[0] = xmin
    X[-1] = xmax
    for i in range(1, n - 1):
        X[i] = int(np.exp(lmin + dl * i))
    return X

def random_split(n_sample, r, seed=0):
    """
    Obtain train/test indexes with a test ratio 

    Parameters
    ----------
    n_sample: int
              giving the number of samples
    r: float
       test ratio

    Returns: 
    -------
    train_list: list
                train indexes
    test_list: list
               test indexes
    """

    if r==0.0:
        return range(n_sample), []

    np.random.seed(seed)

    all_list = np.arange(n_sample)
    randomchoice = np.random.rand(n_sample)
    test_member_mask = (randomchoice < r)
    train_list = all_list[~test_member_mask]
    test_list = all_list[test_member_mask]

    if len(test_list) < 1:
        raise ValueError("No test set selected. Increase sample size and/or test ratio.")

    return train_list, test_list

def kernel_random_split(X, y, r=0.05, seed=0):
    """

    Parameters
    ----------
    X: array-like, shape=[n_samples,n_desc]
       kernel matrix
    y: array-like, shape=[n_samples]
       labels
    r: float
       test ratio

    Returns
    -------
    X_train, X_test: np.matrix
                 train/test kernel matrix
    y_train, y_test: np.array
                 train/test labels
    train_list, test_list: list
                 train/test indexes
    """

    np.random.seed(seed)

    if X.shape[0] != X.shape[1]:
        raise ValueError('Kernel matrix is not square')
    if len(X) != len(y):
        raise ValueError('Length of the vector of properties is not the same as the number of samples')

    n_sample = len(X)
    all_list = np.arange(n_sample)
    randomchoice = np.random.rand(n_sample)
    test_member_mask = (randomchoice < r)
    train_list = all_list[~test_member_mask]
    test_list = all_list[test_member_mask]

    X_train = X[:, train_list][train_list]
    y_train = y[train_list]

    X_test = X[:, train_list][test_list]
    y_test = y[test_list]

    if len(X_test) < 1:
        raise ValueError("No test set selected. Increase sample size and/or test ratio.")

    return X_train, X_test, y_train, y_test, train_list, test_list


"""
adapted from Felix Musil ML_tools
"""


class KFold(_KFold):
    def __init__(self, n_splits=3, shuffle=False, random_state=None):
        super(KFold, self).__init__(n_splits, shuffle, random_state)

    def get_params(self):
        params = dict(n_splits=self.n_splits, shuffle=self.shuffle, random_state=self.random_state)
        return params


class ShuffleSplit(_ShuffleSplit):
    def __init__(self, n_splits=10, test_size="default", train_size=None, random_state=None):
        super(ShuffleSplit, self).__init__(n_splits, test_size, train_size, random_state)

    def get_params(self):
        params = dict(n_splits=self.n_splits, test_size=self.test_size,
                      train_size=self.train_size, random_state=self.random_state)
        return params


class LCSplit():
    def __init__(self, cv, n_repeats=[10], train_sizes=[10], test_size=None, random_state=None, **cvargs):
        if not isinstance(n_repeats, collections.Iterable) or not isinstance(train_sizes, collections.Iterable):
            raise ValueError("Number of repetitions or training set sizes must be an iterable.")

        if len(n_repeats) != len(train_sizes):
            raise ValueError("Number of repetitions must be equal to length of training set sizes.")

        if any(key in cvargs for key in ('random_state', 'shuffle')):
            raise ValueError("cvargs must not contain random_state or shuffle.")

        self.cv = cv
        self.n_repeats = n_repeats
        self.train_sizes = train_sizes
        self.random_state = random_state
        self.cvargs = cvargs
        self.test_size = test_size
        self.n_splits = np.sum(n_repeats)

    def get_params(self):
        params = dict(cv=self.cv.get_params(), n_repeats=self.n_repeats, train_sizes=self.train_sizes,
                      test_size=self.test_size, random_state=self.random_state, cvargs=self.cvargs)
        return params

    def split(self, X, y=None, groups=None):
        """Generates indices to split data into training and test set.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like, of length n_samples
            The target variable for supervised learning problems.
        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set.
        Returns
        -------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """

        rng = check_random_state(self.random_state)

        for n_repeat, train_size in zip(self.n_repeats, self.train_sizes):
            cv = self.cv(random_state=rng, n_splits=n_repeat, test_size=self.test_size, train_size=train_size,
                         **self.cvargs)
            for train_index, test_index in cv.split(X, y, groups):
                yield train_index, test_index

    def get_n_splits(self, X=None, y=None, groups=None):
        """Returns the number of splitting iterations in the cross-validator
        Parameters
        ----------
        X : object
            Always ignored, exists for compatibility.
            ``np.zeros(n_samples)`` may be used as a placeholder.
        y : object
            Always ignored, exists for compatibility.
            ``np.zeros(n_samples)`` may be used as a placeholder.
        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set.
        Returns
        -------
        n_splits : int
            Returns the number of splitting iterations in the cross-validator.
        """
        rng = check_random_state(self.random_state)
        n_splits = 0
        for n_repeat, train_size in zip(self.n_repeats, self.train_sizes):
            cv = self.cv(random_state=rng, n_splits=n_repeat, test_size=self.test_size, train_size=train_size,
                         **self.cvargs)
            n_splits += cv.get_n_splits(X, y, groups)
        return n_splits



"""
Wrappers to do plots
"""

import numpy as np
from matplotlib import pyplot as plt

from .plot_styles import *

class Plotters:
    """
    Object handing the making of plots.
    Notice that we make all the plots on the same graph.
    A good way of using this is to
    1. make a scatter plot
    2. add clustering results (if any)
    3. add tags to annotate important points

    """

    def __init__(self, fig_spec_dict={}):
        """
        Object handing the making of plots
        Parameters
        ----------
        fig_spec_dict: dictionaries that specify plot options
        e.g.
        fig_spec_dict = {
        'outfile': None,
        'show': True,
        'title': None,
        'xlabel': 'Principal Axis 1',
        'ylabel': 'Principal Axis 2',
        'xaxis': True,  'yaxis': True,
        'remove_tick': False,
        'rasterized': True,
        'fontsize': 16,
        'components':{ 
            "first_p": {"type": 'scatter', ...},
            "second_p": {"type": 'annotate', ...},
            "third_p": {"type": 'cluster', ...}
             }
        }

        Notice that we make all the plots on the same graph.
        A good way of using this is to 
        1. make a scatter plot
        2. add clustering results (if any)
        3. add tags to annotate important points
        """

        # required
        try:
            self.p_spec_dict = fig_spec_dict['components']
        except:
            raise ValueError("Didn't specify any options to do plots. Use `components` key to do so.")

        self.fig_spec = {
        'size': [16, 8],
        'outfile': None,
        'show': True,
        'title': None,
        'xlabel': 'Principal Axis 1',
        'ylabel': 'Principal Axis 2',
        'xaxis': True,  'yaxis': True,
        'remove_tick': False,
        'rasterized': True,
        'fontsize': 16,
        'cmap': 'gnuplot'
        }

        # fill in the values
        for k, v in fig_spec_dict.items():
            if k in self.fig_spec.keys():
                self.fig_spec[k] = v

        # pass down some key information, so all the components for the plot are consistent
        for element in self.p_spec_dict.keys():
            self.p_spec_dict[element]['cmap'] = self.fig_spec['cmap']
            self.p_spec_dict[element]['rasterized'] = self.fig_spec['rasterized']

        # list of plotting objects
        self.engines = {}
        self.acronym = ""

        # intialize plot
        set_nice_font()
        self.fig, self.ax = plt.subplots()
        # size
        self.fig.set_size_inches(self.fig_spec['size'][0], self.fig_spec['size'][1])
        # titles
        if self.fig_spec['xlabel'] is not None:
            self.ax.set_xlabel(self.fig_spec['xlabel'], fontsize=self.fig_spec['fontsize'], labelpad=-3)
        if self.fig_spec['ylabel'] is not None:
            self.ax.set_ylabel(self.fig_spec['ylabel'], fontsize=self.fig_spec['fontsize'], labelpad=-3)
        if self.fig_spec['title'] is not None:
            self.ax.set_title(self.fig_spec['title'], fontsize=self.fig_spec['fontsize'])

        self.bind()

    def add(self, p_spec, tag):
        """
        adding the specifications of a new kernel function
        Parameters
        ----------
        p_spec: dictionary
                specify which atomic descriptor to use 
        """
        self.p_spec_dict[tag] = p_spec

    def pack(self):
        return json.dumps(self.p_spec_dict, sort_keys=True, cls=NpEncoder)

    def get_acronym(self):
        if self.acronym == "":
            for element in self.p_spec_dict.keys():
                self.acronym += self.engines[element].get_acronym()
        return self.acronym

    def bind(self):
        """
        binds the objects that actually make the plots
        these objects need to have .create() method to compute 
        """
        # clear up the objects
        self.engines = {}
        for element in self.p_spec_dict.keys():
            self.engines[element] = self._call(self.p_spec_dict[element])
            self.p_spec_dict[element]['acronym'] = self.engines[element].get_acronym()

    def _call(self, p_spec):
        """
        call the specific kernel objects
        """
        if "type" not in p_spec.keys():
            raise ValueError("Did not specify the type of the kernel function.")
        if p_spec["type"] == "scatter":
            return Plot_Function_Scatter(p_spec)
        if p_spec["type"] == "annotate":
            return Plot_Function_Annotate(p_spec)
        if p_spec["type"] == "cluster":
            return Plot_Function_Cluster(p_spec)
        else:
            raise NotImplementedError 

    def plot(self, X, colors=[], labels=[], tags=[]):
        """Plots a 2D density map given 2D coordinates X
        and properties 
        and tags 
        for every data point.

        Parameters
        ----------
        X : array-like, shape=[n_samples,2]
            Input points.
        colors : array-like, float or integer, shape=[n_samples]
            properties used to color each point
        labels : array-like, float or integer, shape=[n_samples]
            additional properties for each point
        tags : array-like, str, shape=[n_samples]
            tags used to annotate selected points
        """

        for element in self.p_spec_dict.keys():
            self.fig, self.ax = self.engines[element].create(self.fig, self.ax, X, colors, labels, tags)

        # touch ups
        if self.fig_spec['remove_tick']:
            self.ax.tick_params(axis='both', which='both', bottom=False, top=False, 
                               labelbottom=False, right=False, left=False, labelleft=False)
        if self.fig_spec['xaxis'] is not True:
            self.ax.set_xticklabels([])
        if self.fig_spec['yaxis'] is not True:
            self.ax.set_yticklabels([])

        # save and show
        if self.fig_spec['outfile'] is not None:
            self.fig.savefig(self.fig_spec['outfile'])
        if self.fig_spec['show']:
            plt.show()

class Plot_Function_Base:
    def __init__(self, p_spec):
        self.acronym = ""
        pass
    def get_acronym(self):
        # we use an acronym for each plot, so it's easy to find it and refer to it
        return self.acronym
    def create(self, fig, ax, X, colors, labels, tags):
        pass

class Plot_Function_Annotate(Plot_Function_Base):
    def __init__(self, p_spec):
        self.acronym = "annotate"

        self.p_spec = {
        'adtext': True,
        'marker': '^',
        'markersize': 10,
        'markercolor': 'black',
        'textsize': 12,
        'textcolor': 'red' # we can add more options for the adtext part
        }

        # fill in the values
        for k, v in p_spec.items():
            if k in self.p_spec.keys():
                self.p_spec[k] = v

        print("Using annotation plot ...")

    def create(self, fig, ax, X, z=[], labels=[], tags=[]):
        """
        annotate samples using tags.
        Parameters
        ----------
        X : array-like, shape=[n_samples,2]
        Input points.
        tags : array-like, str, shape=[n_samples]
        tags for each point.
        labels and z are not used for this plot style
        """
        texts = []
        for i in range(len(tags)):
            if tags[i] != 'None' and tags[i] != 'none' and tags[i] != '':
                ax.scatter(X[i, 0], X[i, 1], 
                           marker=self.p_spec['marker'], 
                           c=self.p_spec['markercolor'])
                texts.append(ax.text(X[i, 0], X[i, 1], tags[i],
                                     ha='center', va='center', 
                                     fontsize=self.p_spec['textsize'],
                                     color=self.p_spec['textcolor']))
        if self.p_spec['adtext']:
            """ adjust the position of the annotated text, so they don't overlap """
            from adjustText import adjust_text
            adjust_text(texts, on_basemap=True,  # only_move={'points':'', 'text':'x'},
                        expand_text=(1.01, 1.05), expand_points=(1.01, 1.05),
                        force_text=(0.03, 0.5), force_points=(0.01, 0.25),
                        ax=ax, precision=0.01,
                        arrowprops=dict(arrowstyle="-", color='black', lw=1, alpha=0.8))

        return fig, ax

class Plot_Function_Scatter(Plot_Function_Base):
    def __init__(self, p_spec):

        self.acronym = "scatter"

        self.p_spec = {
        'psize': None,
        'rasterized': True,
        'fontsize': 12,
        'cmap': 'gnuplot',
        'alpha': 1.0, # color transparency
        'clabel': None, # label of the colorbar
        'cbar_format': None, #'%1.1f',
        'use_perc': False, # mark the top/bottom ourliers
        'outlier_top_fraction': 0.05, # the fraction of the top ourliers
        'outlier_top_color': 'yellow', # color used to make the top ourliers
        'outlier_bottom_fraction': 0.05, # the fraction of the bottom ourliers
        'outlier_bottom_color': 'black', # color used to make the bottom ourliers
        'vmax': None,
        'vmin': None
        }

        # fill in the values
        for k, v in p_spec.items():
            if k in self.p_spec.keys():
                self.p_spec[k] = v

        print("Using scatter plot ...")

        self.cb = None

    def create(self, fig, ax, X, z, labels=[], tags=[]):
        """
        Plots a 2D scatter map given x,y coordinates and a color for
        every data point

        Parameters
        ----------
        X : array-like, shape=[n_samples,2]
        Input points.
        z : array-like, shape=[n_samples]
        color for each point.
        labels and tags are not used for this plot style
        """

        x, y = X[:, 0], X[:, 1]
        z = np.asarray(z)

        # automatically adjust the marker size according the number of samples
        if self.p_spec['psize'] == None:
            psize = 200*200/len(X)
        else:
            psize = self.p_spec['psize']

        if self.p_spec['use_perc']:
            n_sample = len(x)
            argz = np.argsort(z)
            n_bot_outliers = int(n_sample*self.p_spec['outlier_bottom_fraction'])
            n_top_outliers = int(n_sample*self.p_spec['outlier_top_fraction'])
            bot_outliers = argz[:n_bot_outliers]
            top_outliers = argz[-n_top_outliers:]
            typical = argz[n_bot_outliers:-top_outliers]
            # plot typical
            axscatter = ax.scatter(x[typical], y[typical], c=z[typical], 
                               cmap=self.p_spec['cmap'], 
                               s=psize, 
                               alpha=self.p_spec['alpha'],
                               rasterized=self.p_spec['rasterized'])
            # plot bot outliers
            ax.scatter(x[bot_outliers], y[bot_outliers], c=self.p_spec['outlier_bottom_color'], 
                               cmap=self.p_spec['cmap'], 
                               s=psize, 
                               alpha=self.p_spec['alpha'],
                               rasterized=self.p_spec['rasterized'])
            # plot top outliers
            ax.scatter(x[top_outliers], y[top_outliers], c=self.p_spec['outlier_top_color'], 
                               cmap=self.p_spec['cmap'], 
                               s=psize, 
                               alpha=self.p_spec['alpha'],
                               rasterized=self.p_spec['rasterized'])
        else:
            # check if the labels are discrete
            discrete_label = True
            for iz in z: 
                if not np.equal(np.mod(iz, 1), 0): discrete_label = False
            if discrete_label:
                print("Use discrete colormap ......")
                cmap = plt.cm.get_cmap(self.p_spec['cmap'], int(np.max(z)-np.min(z)) + 1)
                vmin = np.min(z) - 0.5
                vmax = np.max(z) + 0.5
            else:
                cmap = self.p_spec['cmap']
                vmin = self.p_spec['vmin']
                vmax = self.p_spec['vmax']

            axscatter = ax.scatter(x, y, c=z, 
                               cmap=cmap, 
                               s=psize, 
                               alpha=self.p_spec['alpha'],
                               rasterized=self.p_spec['rasterized'],
                               vmax=vmax,
                               vmin=vmin)

        if self.p_spec['cbar_format'] is None:
            color_spread = np.nanmax(z) - np.nanmin(z)
            if color_spread > 2:
                self.p_spec['cbar_format'] = '%d'
            elif color_spread > 0.1:
                self.p_spec['cbar_format'] = '%1.1f'
            else:
                self.p_spec['cbar_format'] = '%1.1e'
        if self.p_spec['clabel'] is not None and self.cb is None: 
            self.cb = fig.colorbar(axscatter, format=self.p_spec['cbar_format'])
            self.cb.ax.locator_params(nbins=5)
        if self.p_spec['clabel'] is not None:
            self.cb.set_label(label=self.p_spec['clabel'], labelpad=1)

        return fig, ax


class Plot_Function_Cluster(Plot_Function_Base):
    """
    Plots a 2D clustering plot given x,y coordinates and a label z for
    every data point.
    Basically we draw a circle centered arround the mean position of the samples 
    belonging to each cluster,
    with a size propotional to log(cluster_size)
    """
    def __init__(self, p_spec):

        self.acronym = "cluster"

        self.p_spec = {
        'w_label': False,
        'circle_size': 20, 
        'facecolor': 'none',
        'edgecolor': 'gray',
        'fontsize': 16,
        'cmap': 'gnuplot',
        'alpha': 1.0 # color transparency
        }

        # fill in the values
        for k, v in p_spec.items():
            if k in self.p_spec.keys():
                self.p_spec[k] = v

        print("Using cluster plot ...")

    def create(self, fig, ax, X, z=[], y=[], tags=[]):
        """
        Parameters
        ----------
        X : array-like, shape=[n_samples,2]
        Input points.
        y : array-like, shape=[n_samples]
        label for every point
        z and tags are not used for this plot style
        """

        # get the cluster size and mean position
        from ..cluster import get_cluster_size, get_cluster_properties
        y_unique = np.unique(y)
        [_, cluster_mx] = get_cluster_properties(y, X[:, 0], 'mean')
        [_, cluster_my] = get_cluster_properties(y, X[:, 1], 'mean')
        [_, cluster_size] = get_cluster_size(y)
        s = {}
        for k in y_unique:
            s[k] = np.log(cluster_size[k]) # default is using log(frequency)

        for k in y_unique:
            ax.plot(cluster_mx[k], cluster_my[k], 'o', 
                    markerfacecolor=self.p_spec['facecolor'],
                    markeredgecolor=self.p_spec['edgecolor'], 
                    markersize=self.p_spec['circle_size'] * s[k])

        if self.p_spec['w_label'] is True:
            for k in y_unique:
                # Position of each label.
                txt = ax.annotate(str(k), xy=(cluster_mx[k], cluster_my[k]),
                              xytext=(0, 0), textcoords='offset points',
                              fontsize=self.p_spec['fontsize'], 
                              horizontalalignment='center', verticalalignment='center')
                txt.set_path_effects([
                              PathEffects.Stroke(linewidth=5, foreground='none'),
                              PathEffects.Normal()])

        return fig, ax

        



"""
Methods and functions to perform dimensionality reduction
"""
import numpy as np
import json
from copy import copy
import umap
from sklearn.manifold import TSNE
from sklearn.preprocessing import StandardScaler

from .sparse_kpca import SPARSE_KPCA
from .ml_pca import PCA
from ..io import NpEncoder


class Dimension_Reducers:
    def __init__(self, dreduce_spec_dict={}):
        """
        Object handing the dimensionality reduction of design matrices
        Parameters
        ----------
        dreduce_spec_dict: dictionaries that specify which dimensionality reduction algorithm to use 
        we can perform a list of D-reduction sequentially.
        e.g.
        dreduce_spec_dict = {
        "preprocessing": {"type": 'SCALE', 'parameter': None},
        "reduce2_umap": {"type": 'UMAP', 'parameter': {"n_components": 10, "n_neighbors": 10, "metric":'euclidean'}}
        }
        e.g.
        dreduce_spec_dict = {
        "reduce1_pca": {"type": 'PCA', 'parameter':{"n_components": 50, "scalecenter":True}},
        "reduce2_tsne": {"type": 'TSNE', 'parameter': {"n_components": 10, "perplexity":20}}
        }
        """
        self.dreduce_spec_dict = dreduce_spec_dict
        # list of Atomic_Descriptor objections
        self.engines = {}
        #self.acronym = "" 
        self._fitted = False
        self.bind()

    def add(self, dreduce_spec, tag):
        """
        adding the specifications of a new dimensionality reducer
        Parameters
        ----------
        dreduce_spec: a dictionary that specify which dimensionality reducer to use 
        """
        self.dreduce_spec_dict[tag] = dreduce_spec

    def pack(self):
        return json.dumps(self.dreduce_spec_dict, sort_keys=True, cls=NpEncoder)
    """
    def get_acronym(self):
        if self.acronym == "":
            for element in self.dreduce_spec_dict.keys():
                self.acronym += self.engines[element].get_acronym()
        return self.acronym
    """
    def bind(self):
        """
        binds the objects that actually performs dimension reduction
        these objects need to have .fit(desc)/.transform(desc)/.fit_transform(desc) methods to perform reduction on a design matrix desc
        """
        # clear up the objects
        self.engines = {}
        for element in self.dreduce_spec_dict.keys():
            self.engines[element] = self._call(self.dreduce_spec_dict[element])
            #self.dreduce_spec_dict[element]['acronym'] = self.engines[element].get_acronym()

    def _call(self, dreduce_spec):
        """
        call the specific descriptor objects
        """
        if "type" not in dreduce_spec.keys():
            raise ValueError("Did not specify the type of the dimensionality reducer.")
        elif dreduce_spec["type"] == "SCALE":
            print ("Using standard scaling of the data ...")
            return StandardScaler()
        elif dreduce_spec["type"] == "PCA":
            print ("Using PCA ...")
            return PCA(**dreduce_spec['parameter'])
        elif dreduce_spec["type"] == "SPARSE_KPCA":
            print ("Using kernel PCA (sparsified) ...")
            return SPARSE_KPCA(**dreduce_spec['parameter'])
        elif dreduce_spec["type"] == "UMAP":
            """https://umap-learn.readthedocs.io/en/latest/api.html#umap.umap_.UMAP.fit_transform"""
            print ("Using UMAP ...")
            return umap.UMAP(**dreduce_spec['parameter'])
        elif dreduce_spec["type"] == "TSNE":
            print ("Using t-sne ...")
            """https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html"""
            return TSNE(**dreduce_spec['parameter'])
        else:
            raise NotImplementedError 

    def fit(self, X):
        """
        compute the global descriptor vector for a frame from atomic contributions
        Parameters
        ----------
        X : array-like, shape=[n_samples,n_dim_high]
            Input points.
        """
        for element in self.dreduce_spec_dict.keys():
            X = self.engines[element].fit(X)
        self._fitted = True

    def fit_transform(self, X):
        """
        compute the global descriptor vector for a frame from atomic contributions
        Parameters
        ----------
        X : array-like, shape=[n_samples,n_dim_high]
            Input points.

        Returns
        -------
        x: array-like, shape=[n_samples,n_dim_low]
        """
        X = X.copy()
        for element in self.dreduce_spec_dict.keys():
            X = self.engines[element].fit_transform(X)
        self._fitted = True
        return X

    def transform(self, X):
        """
        compute the global descriptor vector for a frame from atomic contributions
        Parameters
        ----------
        X : array-like, shape=[n_samples,n_dim_high]
            Input points.

        Returns
        -------
        x: array-like, shape=[n_samples,n_dim_low]
        """
        if self._fitted == False:
            raise ValueError("Haven't fit the dimensionality reducer. Use .fit or .fit_transform first")
        X = X.copy()
        for element in self.dreduce_spec_dict.keys():
            if self.dreduce_spec_dict[element]['type'] == 'TSNE':
                raise ValueError("TSNE does not allow out-of-sampling embedding")
            X = self.engines[element].transform(X)
        return X





"""
ASAPXYZ class for handing atomic coordinate input and compute/output
"""
import os
import glob
import json
from yaml import dump as ydump
from yaml import Dumper
import numpy as np
from ase import Atoms
from ase.io import read, write
from tqdm.auto import tqdm
from joblib import Parallel, delayed
import pandas as pd

from ..io import randomString, NpEncoder
from ..descriptors import Atomic_Descriptors, Global_Descriptors


class ASAPXYZ:
    """extended xyz class

    Parameters
    ----------
    fxyz: string_like
         the path to the extended xyz file
    fmat: string_like
           the name of the descriptors in the extended xyz file
    use_atomic_desc: bool
            return the descriptors for each atom, read from the xyz file
    stride: int
           the stride when reading the xyz file
    """

    def __init__(self, fxyz=None, stride=1, periodic=True, fileformat=None):
        # compile a list of matching xyz files
        # in fact they don't strictly need to be xyz format, anything that can be read by ASE is fine
        # a list of possible file formats: https://wiki.fysik.dtu.dk/ase/ase/io/io.html
        if '*' in fxyz:
            self.fxyz = glob.glob(fxyz)
            print("Find matching input files with coordinates: ", self.fxyz)
        else:
            self.fxyz = fxyz
        # essential
        self.stride = stride
        self.periodic = periodic
        if fileformat is not None:
            import ast
            self.fileformat = ast.literal_eval(fileformat)
        else:
            self.fileformat = {}

        # store the xyz file
        self.frames = None
        self.nframes = 0
        self.natom_list = []  # number of atoms for each frame
        self.total_natoms = 0  # total number of atoms for all frames
        self.global_species = []  # list of elements contains in all frames

        # record the state of the computation, e.g. which descriptors have been computed
        self.computed_desc_dict = {'data': {'fxyz': fxyz}}
        self.computed_desc_dict = {'descriptors': {}}
        # the conversion between tag of the descriptors and their acronyms
        self.tag_to_acronym = {'global': {}, 'atomic': {}}

        # we make a dictionary to store the computed descriptors
        self.global_desc = {}
        # this is for the atomic ones
        self.atomic_desc = {}

        # try to read the xyz file
        try:
            if isinstance(self.fxyz, (tuple, list)):
                self.frames = []
                for f in self.fxyz:
                    self.frames += read(f, slice(0, None, self.stride), **self.fileformat)
            else:
                self.frames = read(self.fxyz, slice(0, None, self.stride), **self.fileformat)
        except:
            raise ValueError('Exception occurred when loading the input file')

        self.nframes = len(self.frames)
        all_species = []
        for i, frame in enumerate(self.frames):
            # record the total number of atoms
            self.natom_list.append(len(frame.get_positions()))
            all_species.extend(frame.get_atomic_numbers())
            if not self.periodic or not np.sum(frame.get_cell()) > 0:
                frame.set_pbc([False, False, False])
            # we also initialize the descriptor dictionary for each frame
            self.global_desc[i] = {}
            self.atomic_desc[i] = {}

        self.total_natoms = np.sum(self.natom_list)
        self.max_atoms = max(self.natom_list)
        # Keep things in plain python for serialisation
        self.global_species = np.unique(all_species).tolist()
        print('load xyz file: ', self.fxyz,
              ', a total of ', str(self.nframes), 'frames',
              ', a total of ', str(self.total_natoms), 'atoms',
              ', with elements: ', self.global_species, '.')

    def get_xyz(self):
        return self.frames

    def get_num_frames(self):
        return self.nframes

    def get_total_natoms(self):
        return self.total_natoms

    def get_natom_list(self):
        return self.natom_list

    def get_global_species(self):
        return self.global_species

    def get_natom_list_by_species(self, species_name=None):
        if species_name is None:
            return self.natom_list
        elif species_name in self.global_species:
            return [a.get_atomic_numbers().tolist().count(species_name) for a in self.frames]
        else:
            raise ValueError("Cannot find the specified chemical species in the data set.")

    def save_state(self, filename, mode='yaml'):
        if mode == 'yaml':
            with open(filename + '-state.yaml', 'w') as yd:
                ydump(self.computed_desc_dict, yd, sort_keys=True, Dumper=Dumper)
        else:
            with open(filename + '-state.json', 'w') as jd:
                json.dump(self.computed_desc_dict, jd, sort_keys=True, cls=NpEncoder)

    def save_descriptor_acronym_state(self, filename, mode='yaml'):

        if mode == 'yaml':
            with open(filename + '-descriptor-acronyms.yaml', 'w') as yd:
                ydump(self.tag_to_acronym, yd, sort_keys=True)
        else:
            with open(filename + '-descriptor-acronyms.json', 'w') as jd:
                json.dump(self.tag_to_acronym, jd, sort_keys=True, cls=NpEncoder)

    def symmetrise(self, sbs=[], symprec=1e-2):
        import spglib
        if len(sbs) == 0:
            sbs = range(self.nframes)
        for i in sbs:
            frame = self.frames[i]
            frame.info['space_group'] = spglib.get_spacegroup(frame, symprec=symprec)

    def standardize(self, sbs=[], symprec=1e-2):
        """
        reduce to primitive cell
        """
        import spglib
        if len(sbs) == 0:
            sbs = range(self.nframes)
        for i in sbs:
            frame = self.frames[i]
            lattice, scaled_positions, numbers = spglib.standardize_cell(frame,
                                                                         to_primitive=1,
                                                                         no_idealize=1,
                                                                         symprec=symprec)
            self.frames[i] = Atoms(numbers=numbers, cell=lattice, scaled_positions=scaled_positions,
                                   pbc=frame.get_pbc())

    def _add_info_to_desc_spec(self, desc_spec_dict):
        """
        add some system specific information to the list to descriptor specifications
        Parameters
        ----------
        desc_spec_dict: dictionaries 
                        dict that specify which global descriptor to use.
        """
        for element in desc_spec_dict.keys():
            desc_spec_dict[element]['species'] = self.global_species
            desc_spec_dict[element]['periodic'] = self.periodic
            desc_spec_dict[element]['max_atoms'] = self.max_atoms

    def compute_atomic_descriptors(self, desc_spec_dict={}, sbs=[], tag=None, n_process=1):
        """
        compute the atomic descriptors for selected frames
        Parameters
        ----------
        desc_spec: a list of dictionaries
                 contrains infos on the descriptors to use
        e.g.
        .. code-block:: python

            atomic_desc_dict = {
            "firstsoap": 
            {"type": 'SOAP',"species": [1, 6, 7, 8], "cutoff": 2.0, "atom_gaussian_width": 0.2, "n": 4, "l": 4}}
        
        sbs: array, integer
             the index of the subset of structures to compute
        """

        if len(sbs) == 0:
            sbs = range(self.nframes)
        if tag is None: tag = randomString(6)

        # add some system specific information to the list to descriptor specifications
        self._add_info_to_desc_spec(desc_spec_dict)

        # business!
        atomic_desc = Atomic_Descriptors(desc_spec_dict)

        # serial computation
        if n_process == 1:
            for i in tqdm(sbs):
                frame = self.frames[i]
                # compute atomic descriptor
                self.atomic_desc[i].update(atomic_desc.compute(frame))
        # parallel computation
        elif n_process >= 2:
            results = Parallel(n_jobs=n_process, verbose=1)(delayed(atomic_desc.compute)(self.frames[i]) for i in sbs)
            for i, atomic_desc_dict_now in enumerate(results):
                self.atomic_desc[i].update(atomic_desc_dict_now)
        else:
            raise ValueError("Please set the number of processes to be a positive integer.")

        # we mark down that this descriptor has been computed
        self.computed_desc_dict[tag] = atomic_desc.desc_spec_dict

    def compute_global_descriptors(self, desc_spec_dict={}, sbs=[], keep_atomic=False, tag=None, n_process=1):
        """
        compute the atomic descriptors for selected frames
        Parameters
        ----------
        desc_spec_dict: dictionaries that specify which global descriptor to use.

        e.g.
        .. code-block:: python

            {'global_desc1': 
                          {"type": 'CM'}}

            # or

            {'global_desc2': 
                          {'atomic_descriptor': 
                                       atomic_desc_dict, 
                           'reducer_function': 
                                       reducer_dict
                          }}
        
            atomic_desc_dict = {
                              "firstsoap": 
                                       {"type": 'SOAP',
                                       "species": [1, 6, 7, 8], 
                                       "cutoff": 2.0, 
                                       "atom_gaussian_width": 0.2, 
                                       "n": 4, 
                                       "l": 4}}
        
            reducer_dict = {'first_reducer': 
                                     {'reducer_type': reducer_type,  
                                     'zeta': zeta,
                                     'species': species,
                                     'element_wise': element_wise}}
        
        sbs: array, integer
             list of the indexes of the subset
        """

        if len(sbs) == 0:
            sbs = range(self.nframes)
        if tag is None: tag = randomString(6)

        # add some system specific information to the list to descriptor specifications
        self._add_info_to_desc_spec(desc_spec_dict)

        # business! Intialize a Global_Descriptors object
        global_desc = Global_Descriptors(desc_spec_dict)

        # serial computation
        if n_process == 1:
            for i in tqdm(sbs):
                frame = self.frames[i]
                # compute atomic descriptor
                desc_dict_now, atomic_desc_dict_now = global_desc.compute(frame)
                self.global_desc[i].update(desc_dict_now)
                if keep_atomic:
                    for _, v in atomic_desc_dict_now.items():
                        self.atomic_desc[i].update(v)
        # parallel computation
        elif n_process >= 2:
            results = Parallel(n_jobs=n_process, verbose=1)(delayed(global_desc.compute)(self.frames[i]) for i in sbs)
            for i, (desc_dict_now, atomic_desc_dict_now) in enumerate(results):
                self.global_desc[i].update(desc_dict_now)
                if keep_atomic:
                    for _, v in atomic_desc_dict_now.items():
                        self.atomic_desc[i].update(v)
        else:
            raise ValueError("Please set the number of processes to be a positive integer.")

        # we mark down that this descriptor has been computed
        self.computed_desc_dict['descriptors'][tag] = global_desc.desc_spec_dict

    def fetch_computed_descriptors(self, desc_dict_keys=[], sbs=[]):
        """
        Fetch the computed descriptors for selected frames
        Parameters
        ----------
        desc_spec_keys: a list (str-like) of keys 
                    for which computed descriptors to fetch.
        sbs: array, integer

        Returns
        -------
        desc: np.matrix [n_frame, n_desc]
        """
        if len(sbs) == 0:
            sbs = list(range(self.nframes))
        return np.vstack([self._parse_computed_descriptors_singleframe(desc_dict_keys, i) for i in sbs])

    def fetch_computed_atomic_descriptors(self, desc_dict_keys=[], sbs=[]):
        """
        Fetch the computed atomic descriptors for selected frames
        Parameters
        ----------
        desc_spec_keys: a list (str-like) of keys
                    for which computed descriptors to fetch.
        sbs: array, integer

        Returns
        -------
        desc: np.matrix [n_atoms, n_desc]
        """
        if len(sbs) == 0:
            sbs = list(range(self.nframes))
        return np.vstack([self._parse_computed_atomic_descriptors_singleframe(desc_dict_keys, i) for i in sbs])

    def _parse_computed_descriptors_singleframe(self, desc_dict_keys=[], i=0):
        """return the global descriptor computed for frame i"""
        # TODO: use the nested dictionary search `extract_from_nested_dict` in ..io
        desc_array = np.array([])
        for e in desc_dict_keys:
            try:
                desc_array = np.append(desc_array, self.global_desc[i][e]['descriptors'])
            except:
                # if we use atomic to global descriptor, this is a nested dictionary
                for e2 in self.global_desc[i][e].keys():
                    for e3 in self.global_desc[i][e][e2].keys():
                        desc_array = np.append(desc_array, self.global_desc[i][e][e2][e3]['descriptors'])
        return desc_array

    def _parse_computed_atomic_descriptors_singleframe(self, desc_dict_keys=[], i=0):
        """return the atomic descriptor computed for frame i"""
        return np.hstack( self.atomic_desc[i][e]['atomic_descriptors'] for e in desc_dict_keys)

    def _write_computed_descriptors_to_xyz(self, desc_dict_now, frame):
        """  
        we recorded the computed descriptors to the xyz object
        we use acronym to record the entry in the extended xyz file, so it's much easier to ready by human
        """
        for e in desc_dict_now.keys():
            try:
                frame.info[desc_dict_now[e]['acronym']] = desc_dict_now[e]['descriptors']
                self.tag_to_acronym['global'][e] = desc_dict_now[e]['acronym']
            except:
                # if we use atomic to global descriptor, this is a nested dictionary
                self.tag_to_acronym['global'][e] = {}
                for e2 in desc_dict_now[e].keys():
                    self.tag_to_acronym['global'][e][e2] = {}
                    for e3 in desc_dict_now[e][e2].keys():
                        frame.info[desc_dict_now[e][e2][e3]['acronym']] = desc_dict_now[e][e2][e3]['descriptors']
                        self.tag_to_acronym['global'][e][e2][e3] = desc_dict_now[e][e2][e3]['acronym']

    def _write_computed_atomic_descriptors_to_xyz(self, atomic_desc_dict_now, frame):
        """  
        we recorded the computed descriptors to the xyz object
        we use acronym to record the entry in the extended xyz file, so it's much easier to ready by human
        """
        for e in atomic_desc_dict_now.keys():
            frame.new_array(atomic_desc_dict_now[e]['acronym'],
                            atomic_desc_dict_now[e]['atomic_descriptors'])
            self.tag_to_acronym['atomic'][e] = atomic_desc_dict_now[e]['acronym']

    def _desc_name_with_wild_card(self, desc_name_list, atomic_desc=False):
        """
        Use a wildcard when specifying the name of the descriptors
        """
        new_desc_name = []
        for desc_name in desc_name_list:
            # print("desc_name", desc_name)
            if desc_name == '*':
                import re
                possible_desc_prefix = ['SOAP', 'ACSF', 'LMBTR', 'FCHL19', 'CM', 'pca', 'skpca', 'umap', 'tsne']
                for pre in possible_desc_prefix:
                    if atomic_desc:
                        for key in self.frames[0].arrays.keys():
                            if re.search(pre + '.+', key):
                                new_desc_name.append(key)
                    else:
                        for key in self.frames[0].info.keys():
                            if re.search(pre + '.+', key):
                                new_desc_name.append(key)
            elif '*' in desc_name:
                import re
                if atomic_desc:
                    for key in self.frames[0].arrays.keys():
                        if re.search(desc_name.replace('*', '.+'), key):
                            new_desc_name.append(key)
                else:
                    for key in self.frames[0].info.keys():
                        if re.search(desc_name.replace('*', '.+'), key):
                            new_desc_name.append(key)
            else:
                new_desc_name.append(desc_name)
        return new_desc_name

    def get_descriptors(self, desc_name_list=[], use_atomic_desc=False, species_name=None):
        """ extract the descriptor array from each frame

        Parameters
        ----------
        desc_name_list: a list of strings
                        the name of the .info[] in the extended xyz file
        use_atomic_desc: bool
                         return the descriptors for each atom, read from the xyz file
        species_name: int
                      the atomic number of the species selected.
                      Only the desciptors of atoms of the specified specied will be returned.
                      species_name=None means all atoms are selected.
                      
        Returns
        -------
        desc: np.matrix
        atomic_desc: np.matrix
        """
        desc = []
        atomic_desc = []

        if isinstance(desc_name_list, str):
            desc_name_list = [desc_name_list]

        desc_name_list = self._desc_name_with_wild_card(desc_name_list)
        print("Find the following descriptor names that match the specifications: ", desc_name_list)

        # load from xyz file
        try:
            # retrieve the descriptor vectors --- both of these throw a ValueError if any are missing or are of wrong shape
            desc = np.hstack(
                np.vstack([a.info[desc_name] for a in self.frames]) for desc_name in desc_name_list)
            print("Use global descriptor matrix with shape: ", np.shape(desc))
            # get the atomic descriptors with the same name
            if use_atomic_desc:
                atomic_desc = self.get_atomic_descriptors(desc_name_list, species_name)
        except:
            print("Cannot find the specified descriptors from xyz")

        return desc, atomic_desc

    def get_atomic_descriptors(self, desc_name_list=[], species_name=None):
        """ extract the descriptor array from each frame

        Parameters
        ----------
        desc_name_list: a list of strings
                     the name of the .info[] in the extended xyz file
        species_name: int
                        the atomic number of the species selected.
                        Only the desciptors of atoms of the specified specied will be returned.
                    species_name=None means all atoms are selected.
                         
        Returns
        -------
        atomic_desc: np.matrix
        """

        atomic_desc = []

        if isinstance(desc_name_list, str):
            desc_name_list = [desc_name_list]

        desc_name_list = self._desc_name_with_wild_card(desc_name_list, True)
        print("Find the following atomic descriptor names that match the specifications: ", desc_name_list)

        # load from xyz file
        try:
            if species_name is None:
                atomic_desc = np.vstack(
                    np.concatenate([a.get_array(desc_name) for a in self.frames]) for desc_name in desc_name_list)
            elif species_name in self.global_species:
                atomic_desc = np.hstack(np.concatenate(
                    [self._get_atomic_descriptors_by_species(a, desc_name, species_name) for a in self.frames]) for
                                              desc_name in desc_name_list)
            else:
                raise ValueError("Cannot find the specified chemical species in the data set.")
            print("Use atomic descriptor matrix with shape: ", np.shape(atomic_desc))
        except:
            print("Cannot find the specified atomic descriptors from xyz")

        return atomic_desc

    def _get_atomic_descriptors_by_species(self, frame, desc_name, species_name=None):
        species_index = [i for i, s in enumerate(frame.get_atomic_numbers()) if s == species_name]
        return frame.get_array(desc_name)[species_index]

    def get_property(self, y_key=None, extensive=False, sbs=[]):
        """ extract specified property from selected frames

        Parameters
        ----------
        y_key: string_like
               the name of the property in the extended xyz file
        sbs: array, integer

        Returns
        -------
        y_all: array [N_samples]
        """

        if len(sbs) == 0:
            sbs = range(self.nframes)

        y_all = []
        try:
            for i in sbs:
                frame = self.frames[i]
                if y_key == 'volume' or y_key == 'Volume':
                    y_all.append(frame.get_volume() / len(frame.get_positions()))
                elif y_key == 'size' or y_key == 'Size':
                    y_all.append(len(frame.get_positions()))
                elif y_key == 'index' or y_key == 'Index' or y_key == None:
                    y_all.append(i)
                elif isinstance(y_key, int) and int(y_key) in self.global_species:
                    # count the number of atoms of this specified chemical element
                    y_all.append(frame.get_atomic_numbers().tolist().count(int(y_key)))
                elif extensive:
                    y_all.append(frame.info[y_key] / len(frame.get_positions()))
                else:
                    y_all.append(frame.info[y_key])
        except:
            try:
                for i in sbs:
                    frame = self.frames[i]
                    if extensive:
                        # use the sum of atomic properties
                        y_all.append(np.sum(frame.get_array(y_key)))
                    else:
                        # use the average of atomic properties
                        y_all.append(np.mean(frame.get_array(y_key)))
            except:
                raise ValueError('Cannot load the property vector')
        if len(np.shape(y_all)) > 1:
            raise ValueError('The property from the xyz file has more than one column')
        return np.array(y_all)

    def get_atomic_property(self, y_key=None, extensive=False, sbs=[], species_name=None):
        """ extract the property array from each atom

        Parameters
        ----------
        y_key: string_like
             the name of the property in the extended xyz file
        sbs: array, integer
        specie: int
                 the atomic number of the species selected.
                      Only the properties of atoms of the specified specied will be returned.
                      species_name=None means all atoms are selected.
        Returns
        -------
        y_all: array [N_atoms]
        """

        if len(sbs) == 0:
            sbs = range(self.nframes)

        y_all = []
        try:
            # y_all = np.concatenate([a.get_array(y_key) for a in self.frames[sbs]]) # this doesn't work ?!
            for i in sbs:
                frame = self.frames[i]
                if species_name is None:
                    y_all = np.append(y_all, frame.get_array(y_key))
                elif species_name in self.global_species:
                    y_all = np.append(y_all, self._get_atomic_descriptors_by_species(frame, y_key, species_name))
                else:
                    raise ValueError("Cannot find the specified chemical species in the data set.")
        except:
            try:
                for index, y in enumerate(self.get_property(y_key, extensive, sbs)):
                    y_all = np.append(y_all, y * np.ones(self.natom_list[index]))
                print("Cannot find the atomic properties, use the per-frame property instead")
            except:
                raise ValueError('Cannot load the property vector')
        if len(np.shape(y_all)) > 1:
            raise ValueError('The property from the xyz file has more than one column')
        return y_all

    def set_descriptors(self, desc=None, desc_name=None):
        """ write the descriptor array to the atom object

        Parameters
        ----------
        desc: np.matrix, shape=[n_descriptors, n_frames]

        Returns
        -------
        """
        # TODO: (maybe?) rename this into set_info. Need to change pca.py/kpca.py/... that uses this function
        # check if the length of the descriptor matrix is the same as the number of frames
        if len(desc) != self.nframes and self.nframes > 1:
            raise ValueError('The length of the descriptor matrix is not the same as the number of frames.')
        if len(desc) != self.total_natoms and self.nframes == 1:
            raise ValueError('The length of the descriptor matrix is not the same as the number of atoms.')

        if self.nframes > 1:
            for i, frame in enumerate(self.frames):
                frame.info[desc_name] = desc[i]
        else:
            self.frames[0].new_array(desc_name, desc)

    def set_atomic_descriptors(self, atomic_desc=None, atomic_desc_name=None, species_name=None):
        """ write the descriptor array to the atom object

        Parameters
        ----------
        desc: np.matrix, shape=[n_descriptors, n_atoms]

        Returns
        -------
        """
        # check if the length of the descriptor matrix is the same as the total number of atoms
        if species_name is None and len(atomic_desc) != self.total_natoms:
            raise ValueError('The length of the atomic descriptor matrix is not the same as the total number of atoms.')

        atom_index = 0

        if species_name is None:
            for i, frame in enumerate(self.frames):
                natomnow = self.natom_list[i]
                frame.new_array(atomic_desc_name, np.array(atomic_desc)[atom_index:atom_index + natomnow])
                atom_index += natomnow
        else:
            if atomic_desc.ndim == 1:
                n_desc = 1
            else:
                n_desc = np.shape(atomic_desc)[1]
            for i, frame in enumerate(self.frames):
                array_now = np.zeros((self.natom_list[i], n_desc), dtype=float)
                for j, s in enumerate(frame.get_atomic_numbers()):
                    if s == species_name:
                        array_now[j] = atomic_desc[atom_index]
                        atom_index += 1
                    else:
                        array_now[j] = np.nan
                frame.new_array(atomic_desc_name, np.array(array_now))

    def remove_descriptors(self, desc_name_list=[]):
        """
        remove the desciptors
        """
        if isinstance(desc_name_list, str):
            desc_name_list = [desc_name_list]

        desc_name_list = self._desc_name_with_wild_card(desc_name_list)
        print("removing the global descriptors from output xyz with the names: ", desc_name_list)

        for dn in desc_name_list:
            for frame in self.frames:
                if dn in frame.info:
                    del frame.info[dn]
                else:
                    pass
                    # print("Warning: Cannot parse desc_name "+str(dn)+" when remove_descriptors.")

    def remove_atomic_descriptors(self, desc_name_list=[]):
        """
        remove the desciptors
        """
        if isinstance(desc_name_list, str):
            desc_name_list = [desc_name_list]
        desc_name_list = self._desc_name_with_wild_card(desc_name_list, True)
        print("removing the atomic descriptors from output xyz with the names: ", desc_name_list)

        for dn in desc_name_list:
            for frame in self.frames:
                if dn in frame.arrays:
                    del frame.arrays[dn]
                else:
                    pass
                    # print("Warning: Cannot parse desc_name "+str(dn)+" when remove_descriptors.")

    def load_properties(self, filename, header='infer', prefix='X', **kwargs):
        """
        Load properties from a CSV file

        Read in the CSV file and save the columns to the `info` dictionary of the
        frames.

        Parameters
        ----------
        filename: str 
                Name of the CSV file.
        header: int
            Row number of the header. Defaults to use the first row unless explicit
          names for the columns are given
        """
        data = pd.read_csv(filename, header=header, prefix=prefix, **kwargs)

        for frame, (_, row) in zip(self.frames, data.iterrows()):
            frame.info.update(row.to_dict())

    def write(self, filename, sbs=[], save_acronym=False, wrap_output=True):
        """
        write the selected frames or all the frames to a xyz file

        Parameters
        ----------
        filename: str
        sbs: array, integer
        """

        if len(sbs) == 0:
            sbs = range(self.nframes)

        # prepare for the output
        if os.path.isfile(str(filename) + ".xyz"):
            os.rename(str(filename) + ".xyz", "bck." + str(filename) + ".xyz")

        for i in sbs:
            if wrap_output: self.frames[i].wrap()
            self._write_computed_descriptors_to_xyz(self.global_desc[i], self.frames[i])
            self._write_computed_atomic_descriptors_to_xyz(self.atomic_desc[i], self.frames[i])
            write(str(filename) + ".xyz", self.frames[i], append=True)

        # this acronym state file lets us know how the descriptors correspond to the outputs in the xyz file
        if save_acronym:
            self.save_descriptor_acronym_state(filename)

    def write_chemiscope(self, filename, sbs=None, save_acronym=False, cutoff=None, wrap_output=True):
        """
        write the selected frames or all the frames to ChemiScope JSON

        Parameters
        ----------
        filename: str
        sbs: array, integer
        cutoff: float
                generate cutoff for atomic environments, set to None to disable atomic environments
        """

        from asaplib.io.cscope import write_chemiscope_input

        if sbs is None:
            sbs = range(self.nframes)

        # prepare for the output
        if os.path.isfile(str(filename) + ".xyz"):
            os.rename(str(filename) + ".xyz", "bck." + str(filename) + ".xyz")

        for i in sbs:
            if wrap_output: self.frames[i].wrap()
            self._write_computed_descriptors_to_xyz(self.global_desc[i], self.frames[i])
            self._write_computed_atomic_descriptors_to_xyz(self.atomic_desc[i], self.frames[i])

        # this acronym state file lets us know how the descriptors correspond to the outputs in the xyz file
        if save_acronym:
            self.save_descriptor_acronym_state(filename)

        # Disable atomic environments if there isn't any data
        write_chemiscope_input(filename + '.json.gz', self.frames, cutoff=cutoff)

    def write_descriptor_matrix(self, filename, desc_name_list, sbs=[], comment=''):
        """
        write the selected descriptor matrix in a matrix format to file

        Parameters
        ----------
        filename: str
        desc_name_list: a list of str. 
                Name of the properties/descriptors to write
        sbs: array, integer
        comment: str
        """
        if len(sbs) == 0:
            sbs = range(self.nframes)

        desc, _ = self.get_descriptors(desc_name_list, False, sbs)

        if os.path.isfile(str(filename) + ".desc"):
            os.rename(str(filename) + ".desc", "bck." + str(filename) + ".desc")
        np.savetxt(str(filename) + ".desc", desc, fmt='%4.8f', header=comment)

    def write_atomic_descriptor_matrix(self, filename, desc_name, sbs=[], comment=''):
        """
        write the selected descriptor matrix in a matrix format to file

        Parameters
        ----------
        filename: str
        desc_name: str
                Name of the properties/descriptors to write
        sbs: array, integer
        comment: str
        """

        if len(sbs) == 0:
            sbs = range(self.nframes)

        _, atomic_desc = self.get_descriptors(desc_name, True, sbs)

        if os.path.isfile(str(filename) + ".atomic-desc"):
            os.rename(str(filename) + ".atomic-desc", "bck." + str(filename) + "atomic-desc")
        np.savetxt(str(filename) + ".atomic-desc", desc, fmt='%4.8f', header=comment)

    def write_computed_descriptors(self, filename, desc_dict_keys=[], sbs=[], comment=''):
        """
        write the computed descriptors for selected frames
        Parameters
        ----------
        desc_spec_keys: list
              a list (str-like) of keys for which computed descriptors to fetch.
        sbs: array, integer

        Returns
        -------
        desc: np.matrix [n_frame, n_desc]
        """
        if len(sbs) == 0:
            sbs = range(self.nframes)
        np.savetxt(str(filename) + ".desc", self.fetch_computed_descriptors(desc_dict_keys, sbs), fmt='%4.8f',
                   header=comment)



"""
For geting numerical derivatives of a time dependent series
real or complex
input format
[[i, x_i]]
"""

import numpy as np


def gettxv(xt, dt=1):
    # assume the timestep is constant
    txv = np.zeros((len(xt) - 2, 3))
    mx = np.mean(xt[:, 1])
    # print mx
    for i in range(len(xt) - 2):
        txv[i, 0] = xt[i + 1, 0] * dt
        txv[i, 1] = xt[i + 1, 1] - mx
        txv[i, 2] = (xt[i + 2, 1] - xt[i, 1]) / ((xt[i + 2, 0] - xt[i, 0]) * dt)
    return txv


def getcomplextxv(xt, dt=1):
    # assume the timestep is constant
    txv = np.zeros((len(xt) - 2, 3), dtype=np.complex_)
    mx = np.mean(xt[:, 1]) + 1j * np.mean(xt[:, 2])
    # print mx
    for i in range(len(xt) - 2):
        txv[i, 0] = xt[i + 1, 0] * dt
        txv[i, 1] = xt[i + 1, 1] + 1j * xt[i + 1, 2] - mx
        txv[i, 2] = (xt[i + 2, 1] + 1j * xt[i + 2, 1] - xt[i, 1] - 1j * xt[i, 1]) / ((xt[i + 2, 0] - xt[i, 0]) * dt)
    return txv


def getfftxv(fxx):
    fdxx = np.zeros((len(fxx), 2), dtype=np.complex_)
    for i, f in enumerate(fxx):
        fdxx[i, 0] = fxx[i, 0]
        fdxx[i, 1] = 1j * fxx[i, 0] * fxx[i, 1]
    return fdxx



def system_pair_bond_lengths(Zs, length_scales):
    # shortest bond in this composition
    shortest_bond = min([length_scales[str(Z)]["min_bond_len"][0] for Z in Zs])

    # longest bond in this composition
    longest_bond = max([length_scales[str(Z)]["bond_len"][0] for Z in Zs])

    return shortest_bond, longest_bond


def round_sigfigs(v, n_sig_figs):
    # https://stackoverflow.com/questions/3410976/how-to-round-a-number-to-significant-figures-in-python
    return '{:g}'.format(float('{:.{p}g}'.format(v, p=n_sig_figs)))


uni_length_scales = {"1": {"bond_len": [1.2, "NB VASP auto_length_scale"],
                           "min_bond_len": [0.75, "NB VASP auto_length_scale"],
                           "other links": {},
                           "vol_per_atom": [3.4, "NB VASP auto_length_scale"]},
                     "11": {"bond_len": [3.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [35, "NB VASP auto_length_scale"]},
                     "12": {"bond_len": [3.2, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [23, "NB VASP auto_length_scale"]},
                     "13": {"bond_len": [2.9, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.5, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [17, "NB VASP auto_length_scale"]},
                     "14": {"bond_len": [2.4, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.3, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [20, "NB VASP auto_length_scale"]},
                     "15": {"bond_len": [2.5, "NB VASP auto_length_scale"],
                            "min_bond_len": [1.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [15, "NB VASP auto_length_scale"]},
                     "16": {"bond_len": [2.3, "NB VASP auto_length_scale"],
                            "min_bond_len": [1.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [17, "NB VASP auto_length_scale"]},
                     "17": {"bond_len": [2.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [2, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [36, "NB VASP auto_length_scale"]},
                     "19": {"bond_len": [4.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [3.5, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [71, "NB VASP auto_length_scale"]},
                     "20": {"bond_len": [3.8, "NB VASP auto_length_scale"],
                            "min_bond_len": [3, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [41, "NB VASP auto_length_scale"]},
                     "21": {"bond_len": [3.2, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.6, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [22, "NB VASP auto_length_scale"]},
                     "22": {"bond_len": [2.9, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.4, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [17, "NB VASP auto_length_scale"]},
                     "23": {"bond_len": [2.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [1.8, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [13, "NB VASP auto_length_scale"]},
                     "24": {"bond_len": [2.5, "NB VASP auto_length_scale"],
                            "min_bond_len": [1.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [12, "NB VASP auto_length_scale"]},
                     "26": {"bond_len": [2.5, "NB VASP auto_length_scale"],
                            "min_bond_len": [1.8, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [11, "NB VASP auto_length_scale"]},
                     "27": {"bond_len": [2.5, "NB VASP auto_length_scale"],
                            "min_bond_len": [2, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [11, "NB VASP auto_length_scale"]},
                     "28": {"bond_len": [2.5, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [11, "NB VASP auto_length_scale"]},
                     "29": {"bond_len": [2.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.2, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [12, "NB VASP auto_length_scale"]},
                     "3": {"bond_len": [3, "NB VASP auto_length_scale"],
                           "min_bond_len": [2.4, "NB VASP auto_length_scale"],
                           "other links": {},
                           "vol_per_atom": [20, "NB VASP auto_length_scale"]},
                     "30": {"bond_len": [2.8, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.5, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [15, "NB VASP auto_length_scale"]},
                     "31": {"bond_len": [3, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.5, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [19, "NB VASP auto_length_scale"]},
                     "32": {"bond_len": [2.5, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [24, "NB VASP auto_length_scale"]},
                     "33": {"bond_len": [2.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [20, "NB VASP auto_length_scale"]},
                     "34": {"bond_len": [2.8, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.2, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [22, "NB VASP auto_length_scale"]},
                     "35": {"bond_len": [2.8, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.3, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [29, "NB VASP auto_length_scale"]},
                     "37": {"bond_len": [4.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [3.7, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [81, "NB VASP auto_length_scale"]},
                     "38": {"bond_len": [4.2, "NB VASP auto_length_scale"],
                            "min_bond_len": [3.8, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [54, "NB VASP auto_length_scale"]},
                     "39": {"bond_len": [3.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [33, "NB VASP auto_length_scale"]},
                     "4": {"bond_len": [2.2, "NB VASP auto_length_scale"],
                           "min_bond_len": [2.1, "NB VASP auto_length_scale"],
                           "other links": {},
                           "vol_per_atom": [7.9, "NB VASP auto_length_scale"]},
                     "40": {"bond_len": [3.2, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.3, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [23, "NB VASP auto_length_scale"]},
                     "41": {"bond_len": [2.9, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [18, "NB VASP auto_length_scale"]},
                     "42": {"bond_len": [2.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [16, "NB VASP auto_length_scale"]},
                     "43": {"bond_len": [2.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [14, "NB VASP auto_length_scale"]},
                     "44": {"bond_len": [2.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [14, "NB VASP auto_length_scale"]},
                     "45": {"bond_len": [2.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.2, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [14, "NB VASP auto_length_scale"]},
                     "46": {"bond_len": [2.8, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.4, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [15, "NB VASP auto_length_scale"]},
                     "47": {"bond_len": [2.9, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.6, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [18, "NB VASP auto_length_scale"]},
                     "48": {"bond_len": [3.2, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [23, "NB VASP auto_length_scale"]},
                     "49": {"bond_len": [3.4, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.8, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [27, "NB VASP auto_length_scale"]},
                     "5": {"bond_len": [1.8, "NB VASP auto_length_scale"],
                           "min_bond_len": [1.6, "NB VASP auto_length_scale"],
                           "other links": {},
                           "vol_per_atom": [8.3, "NB VASP auto_length_scale"]},
                     "50": {"bond_len": [2.9, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.6, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [36, "NB VASP auto_length_scale"]},
                     "51": {"bond_len": [3.1, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.5, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [29, "NB VASP auto_length_scale"]},
                     "52": {"bond_len": [3.2, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.6, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [32, "NB VASP auto_length_scale"]},
                     "53": {"bond_len": [3.2, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.7, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [39, "NB VASP auto_length_scale"]},
                     "56": {"bond_len": [4.3, "NB VASP auto_length_scale"],
                            "min_bond_len": [3.5, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [59, "NB VASP auto_length_scale"]},
                     "57": {"bond_len": [3.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.6, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [36, "NB VASP auto_length_scale"]},
                     "58": {"bond_len": [3.4, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.6, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [27, "NB VASP auto_length_scale"]},
                     "59": {"bond_len": [3.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.8, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [36, "NB VASP auto_length_scale"]},
                     "6": {"bond_len": [1.4, "NB VASP auto_length_scale"],
                           "min_bond_len": [1.3, "NB VASP auto_length_scale"],
                           "other links": {},
                           "vol_per_atom": [5.7, "NB VASP auto_length_scale"]},
                     "60": {"bond_len": [3.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.6, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [35, "NB VASP auto_length_scale"]},
                     "61": {"bond_len": [3.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.7, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [34, "NB VASP auto_length_scale"]},
                     "62": {"bond_len": [3.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.7, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [33, "NB VASP auto_length_scale"]},
                     "64": {"bond_len": [3.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [33, "NB VASP auto_length_scale"]},
                     "65": {"bond_len": [3.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [32, "NB VASP auto_length_scale"]},
                     "66": {"bond_len": [3.5, "NB VASP auto_length_scale"],
                            "min_bond_len": [3.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [31, "NB VASP auto_length_scale"]},
                     "69": {"bond_len": [3.5, "NB VASP auto_length_scale"],
                            "min_bond_len": [3.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [30, "NB VASP auto_length_scale"]},
                     "7": {"bond_len": [1.6, "NB VASP auto_length_scale"],
                           "min_bond_len": [1.1, "NB VASP auto_length_scale"],
                           "other links": {},
                           "vol_per_atom": [9.2, "NB VASP auto_length_scale"]},
                     "70": {"bond_len": [3.8, "NB VASP auto_length_scale"],
                            "min_bond_len": [3.3, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [39, "NB VASP auto_length_scale"]},
                     "72": {"bond_len": [3.2, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.6, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [22, "NB VASP auto_length_scale"]},
                     "73": {"bond_len": [2.9, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [18, "NB VASP auto_length_scale"]},
                     "74": {"bond_len": [2.8, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [16, "NB VASP auto_length_scale"]},
                     "75": {"bond_len": [2.8, "NB VASP auto_length_scale"],
                            "min_bond_len": [2, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [15, "NB VASP auto_length_scale"]},
                     "76": {"bond_len": [2.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.1, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [14, "NB VASP auto_length_scale"]},
                     "77": {"bond_len": [2.7, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.2, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [15, "NB VASP auto_length_scale"]},
                     "78": {"bond_len": [2.8, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.3, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [16, "NB VASP auto_length_scale"]},
                     "79": {"bond_len": [2.9, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.5, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [18, "NB VASP auto_length_scale"]},
                     "8": {"bond_len": [1.7, "NB VASP auto_length_scale"],
                           "min_bond_len": [1.2, "NB VASP auto_length_scale"],
                           "other links": {},
                           "vol_per_atom": [11, "NB VASP auto_length_scale"]},
                     "80": {"bond_len": [3.3, "NB VASP auto_length_scale"],
                            "min_bond_len": [3, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [28, "NB VASP auto_length_scale"]},
                     "81": {"bond_len": [3.4, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [31, "NB VASP auto_length_scale"]},
                     "82": {"bond_len": [3.5, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.9, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [31, "NB VASP auto_length_scale"]},
                     "83": {"bond_len": [3.3, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.7, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [34, "NB VASP auto_length_scale"]},
                     "84": {"bond_len": [3.3, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.8, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [36, "NB VASP auto_length_scale"]},
                     "89": {"bond_len": [4, "NB VASP auto_length_scale"],
                            "min_bond_len": [3.5, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [46, "NB VASP auto_length_scale"]},
                     "9": {"bond_len": [2.1, "NB VASP auto_length_scale"],
                           "min_bond_len": [1.4, "NB VASP auto_length_scale"],
                           "other links": {},
                           "vol_per_atom": [15, "NB VASP auto_length_scale"]},
                     "90": {"bond_len": [3.6, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.7, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [32, "NB VASP auto_length_scale"]},
                     "91": {"bond_len": [3.3, "NB VASP auto_length_scale"],
                            "min_bond_len": [2.4, "NB VASP auto_length_scale"],
                            "other links": {},
                            "vol_per_atom": [25, "NB VASP auto_length_scale"]}}



"""
Base classes for clustering algorithms.
"""

from sklearn.base import ClusterMixin


class ClusterBase(ClusterMixin):
    """
    Data structure to perform clustering and store data associated with the clustering output.
    """

    def __init__(self):
        pass

    def fit(self, X, y=None):
        """

        Parameters
        ----------
        X
        y

        Returns
        -------

        """
        return self

    def get_cluster_labels(self, index=[]):
        """

        Parameters
        ----------
        index

        Returns
        -------

        """
        pass

    def get_params(self, deep=True):
        """

        Parameters
        ----------
        deep

        Returns
        -------

        """
        pass

    def get_name(self):
        """

        Returns
        -------

        """
        return type(self).__name__


class FitClusterBase(object):
    def __init__(self):
        pass

    def fit(self, dmatrix, rho=None):
        """

        Parameters
        ----------
        dmatrix
        rho

        Returns
        -------

        """
        pass



"""
Methods and functions to compute global desciptors
"""
import numpy as np
import json
from ..io import NpEncoder, randomString
from .atomic_to_global import Atomic_2_Global_Descriptors
from .atomic_descriptors import Atomic_Descriptors


class Global_Descriptors:
    def __init__(self, desc_spec_dict={}):
        """
        Object handing the specification and the computation of global descriptors
        global descriptors mean descriptors of a whole structure
        atomic descriptors mean descriptors of an atom centered environment inside a structure

        Parameters
        ----------
        desc_spec_dict: dictionaries that specify which global descriptor to use.
        We have two options here
        1. Some descriptors are already global in nature, e.g. the Coulomb Matrix, Morgan fingerprints, etc.
           So we can specify them as, e.g.
        {'global_desc2': 
        {"type": 'CM', "max_atoms" 30
        }}

        2. First compute an atomic descriptors (e.g. SOAP, ACSF,...) and convert to global ones
        e.g.
        {'global_desc2': {'atomic_descriptor': atomic_desc_dict, 'reducer_function': reducer_dict}}
        and
        atomic_desc_dict = {
        "firstsoap": 
        {"type": 'SOAP',"species": [1, 6, 7, 8], "cutoff": 2.0, "atom_gaussian_width": 0.2, "n": 4, "l": 4}
        }
        and
        reducer_dict = {'first_reducer': {'reducer_type': reducer_type,  
                          'zeta': zeta,
                          'species': species,
                          'element_wise': element_wise}}
        """

        self.desc_spec_dict = desc_spec_dict
        # list of Atomic_Descriptor objections
        self.engines = {}
        self.acronym = ""

        self.bind()

    def add(self, desc_spec, tag):
        """
        adding the specifications of a new atomic descriptors
        Parameters
        ----------
        desc_spec: a dictionary that specify which atomic descriptor to use 
        """
        self.desc_spec_dict[tag] = desc_spec

    def pack(self):
        return json.dumps(self.desc_spec_dict, sort_keys=True, cls=NpEncoder)

    def bind(self):
        """
        binds the objects that actually compute the descriptors
        these objects need to have .create(frame) method to compute the descriptors of frame (a xyz object)
        """
        # clear up the objects
        self.engines = {}
        for element in self.desc_spec_dict.keys():
            self.engines[element] = self._call(self.desc_spec_dict[element])
            self.desc_spec_dict[element]['acronym'] = self.engines[element].get_acronym()

    def _call(self, desc_spec):
        """
        call the specific descriptor objects
        """
        if "atomic_descriptor" in desc_spec.keys() and "reducer_function" in desc_spec.keys():
            return Global_Descriptor_from_Atomic(desc_spec)
        elif "type" not in desc_spec.keys():
            raise ValueError("Did not specify the type of the descriptor.")
        if desc_spec["type"] == "CM":
            return Global_Descriptor_CM(desc_spec)
        elif desc_spec["type"] == "MORGAN":
            return Global_Descriptor_Morgan(desc_spec)
        else:
            raise NotImplementedError

    def compute(self, frame):
        """
        compute the global descriptor vector and atomic descriptor matrix (if any) for a frame
        Parameters
        ----------
        frame: ASE atom object. Coordinates of a frame.

        Returns
        -------
        desc_dict: a dictionary. each entry contains the essential info of the descriptor, i.e. acronym 
                          and a np.array [N_desc]. Global descriptors for a frame.
                   e.g. {'d1':{ 'acronym': 'XXX', 'descriptors': `a np.array [N_desc]`}}
        atomic_desc_dict : a dictionary. each entry contains the essential info of the descriptor (acronym) 
                          and a np.array [N_desc*N_atoms]. Atomic descriptors for a frame.
                          e.g. {'ad1':{'acronym':'soap-1', 'atomic_descriptors': `a np.array [N_desc*N_atoms]`}}
        """
        global_desc_dict = {}
        atomic_desc_dict = {}
        for element in self.desc_spec_dict.keys():
            global_desc_dict[element], atomic_desc_dict[element] = self.engines[element].create(frame)
        return global_desc_dict, atomic_desc_dict


class Global_Descriptor_Base:
    def __init__(self, desc_spec):
        self._is_atomic = False
        self.acronym = ""
        pass

    def is_atomic(self):
        return self._is_atomic

    def get_acronym(self):
        # we use an acronym for each descriptor, so it's easy to find it and refer to it
        return self.acronym

    def create(self, frame):
        # return the dictionaries for global descriptors and atomic descriptors (if any)
        return {'acronym': self.acronym, 'descriptors': []}, {}


class Global_Descriptor_from_Atomic(Global_Descriptor_Base):
    def __init__(self, desc_spec):
        """
        First compute an atomic descriptors (e.g. SOAP, ACSF,...) and convert to global ones

        Parameters
        ----------
        desc_spec: dictionaries that specify which global descriptor to use.

        e.g.
        {'global_desc2': {'atomic_descriptor': atomic_desc_dict, 'reducer_function': reducer_dict}}
        and
        atomic_desc_dict = {
        "firstsoap": 
        {"type": 'SOAP',"species": [1, 6, 7, 8], "cutoff": 2.0, "atom_gaussian_width": 0.2, "n": 4, "l": 4}
        }
        and
        reducer_dict = {'first_reducer': {'reducer_type': reducer_type,  
                          'zeta': zeta,
                          'species': species,
                          'element_wise': element_wise}}
        """

        self._is_atomic = True

        if "atomic_descriptor" not in desc_spec.keys() or "reducer_function" not in desc_spec.keys():
            raise ValueError("Need to specify both atomic descriptors and reducer functions to used")

        self.atomic_desc_spec = desc_spec['atomic_descriptor']
        self.reducer_spec = desc_spec['reducer_function']

        # pass down some key information
        if 'species' in desc_spec.keys():
            # add some system specific information to the list to descriptor specifications
            for element in self.atomic_desc_spec.keys():
                self.atomic_desc_spec[element]['species'] = desc_spec['species']
            for element in self.reducer_spec.keys():
                self.reducer_spec[element]['species'] = desc_spec['species']
        if 'periodic' in desc_spec.keys():
            for element in self.atomic_desc_spec.keys():
                self.atomic_desc_spec[element]['periodic'] = desc_spec['periodic']

        # initialize a Atomic_Descriptors object
        self.atomic_desc = Atomic_Descriptors(self.atomic_desc_spec)
        # initialize a Atomic_2_Global_Descriptors object
        self.atomic_2_global = Atomic_2_Global_Descriptors(self.reducer_spec)

        # print("Using Atomic_2_Global_Descriptors ...")

        self.acronym = "atomic-to-global-" + randomString(6)

    def pack(self):
        return {'atomic_descriptor': self.atomic_desc.pack(), 'reducer_function': atomic_2_global.pack()}

    def create(self, frame):
        """
        compute the global descriptor vector for a frame from atomic contributions
        Parameters
        ----------
        frame: ASE atom object. Coordinates of a frame.

        Returns
        -------
        desc_dict: a dictionary. each entry contains the essential info of the descriptor, i.e. acronym 
                          and a np.array [N_desc]. Global descriptors for a frame.
                   e.g. {'d1':{ 'acronym': 'XXX', 'descriptors': `a np.array [N_desc]`}}
        atomic_desc_dict : a dictionary. each entry contains the essential info of the descriptor (acronym) 
                          and a np.array [N_desc*N_atoms]. Atomic descriptors for a frame.
                          e.g. {'ad1':{'acronym':'soap-1', 'atomic_descriptors': `a np.array [N_desc*N_atoms]`}}
        """
        # compute atomic descriptor
        atomic_desc_dict = self.atomic_desc.compute(frame)
        # compute global descriptor for the frame
        return self.atomic_2_global.compute(atomic_desc_dict, frame.get_atomic_numbers()), atomic_desc_dict


class Global_Descriptor_CM(Global_Descriptor_Base):
    def __init__(self, desc_spec):
        """
        make a DScribe CM object
        """

        from dscribe.descriptors import CoulombMatrix

        if "type" not in desc_spec.keys() or desc_spec["type"] != "CM":
            raise ValueError("Type is not CM or cannot find the type of the descriptor")

        # required
        try:
            self.max_atoms = desc_spec['max_atoms']
        except:
            raise ValueError("Not enough information to intialize the `Atomic_Descriptor_CM` object")

        if 'periodic' in desc_spec.keys() and desc_spec['periodic'] == True:
            raise ValueError("Coulomb Matrix cannot be used for periodic systems")

        self.cm = CoulombMatrix(self.max_atoms)
        print("Using CoulombMatrix ...")
        # make an acronym
        self.acronym = "CM" + "-" + str(self.max_atoms)

    def create(self, frame):
        """
        compute the CM descriptor vector for a frame
        Parameters
        ----------
        frame: ASE atom object. Coordinates of a frame.

        Returns
        -------
        desc_dict: a dictionary. each entry contains the essential info of the descriptor, i.e. acronym 
                          and a np.array [N_desc]. Global descriptors for a frame.
                   e.g. {'d1':{ 'acronym': 'CM-*', 'descriptors': `a np.array [N_desc]`}}
        atomic_desc_dict : {}
        """
        if len(frame.get_positions()) > self.max_atoms:
            raise ValueError('the size of the system is larger than the max_atoms of the CM descriptor')
        # notice that we return an empty dictionary for "atomic descriptors"
        return {'acronym': self.acronym, 'descriptors': self.cm.create(frame, n_jobs=1)}, {}


class Global_Descriptor_Morgan(Global_Descriptor_Base):
    def __init__(self, desc_spec):

        if "type" not in desc_spec.keys() or desc_spec["type"] != "MORGAN":
            raise ValueError("Type is not MORGAN or cannot find the type of the descriptor")

        # defaults
        if "length" in desc_spec.keys():
            self.length = desc_spec["length"]
        else:
            self.length = 1024

        if "radius" in desc_spec.keys():
            self.radius = desc_spec["radius"]
        else:
            self.radius = 3

        if 'periodic' in desc_spec.keys() and desc_spec['periodic'] == True:
            raise ValueError("Morgan Fingerprints cannot be used for periodic systems")

        print("Using Morgan Fingerprints ...")
        # make an acronym
        self.acronym = "MORGAN"

    def _get_smiles(self, frame):
        if "smiles" in frame.info:
            # print(frame.info['smiles'])
            return frame.info['smiles']
        elif "SMILES" in frame.info:
            return frame.info['SMILES']
        else:
            raise ValueError('Cannot parse the smile string from the frame.info.')

    def create(self, frame):
        """
        Returns
        -------
        """

        from rdkit.Chem import MolFromSmiles
        from rdkit.Chem.AllChem import GetMorganFingerprintAsBitVect

        smiles = self._get_smiles(frame)
        mol = MolFromSmiles(smiles)
        fps = GetMorganFingerprintAsBitVect(mol, radius=self.radius, nBits=self.length)
        fps = np.array(fps, dtype='float64')

        return {'acronym': self.acronym, 'descriptors': fps}



"""
Methods and functions to convert descriptors to kernels for samples

Kernels are measures of similarity, 
i.e. s(a, b) > s(a, c) if objects a and b are considered “more similar” than objects a and c.
A kernel must also be positive semi-definite.

Essentially, for each pair of samples a and b we compute
k(a,b)
based on the coordinates of descriptors d(a) and d(b) 
"""
import numpy as np
import json
from .ml_kernel_operations import normalizekernel

class Descriptors_to_Kernels:
    def __init__(self, k_spec_dict={}):
        """
        Object handing the specification and the computation of atomic descriptors
        Parameters
        ----------
        k_spec_dict: dictionaries that specify which way to convert descriptors into kernel matrix
        e.g.
        k_spec_dict = {
        "first_kernel": {"type": 'linear', "normalize" = True},
        "second_kernel": {"type": 'cosine'},
        "third_kernel": {"type": 'polynormial', "d":3, "normalize" = True}
        }

        Notice that we can specify multiple kernels here.
        What we do is that:
        1. compute k(a,b) for all these kernel functions
        2. sum up all k(a,b).
        """

        self.k_spec_dict = k_spec_dict
        # list of kernel (similarity measurement) objects
        self.engines = {}
        self.acronym = ""

        self.bind()

    def add(self, k_spec, tag):
        """
        adding the specifications of a new kernel function
        Parameters
        ----------
        k_spec: a dictionary that specify which atomic descriptor to use 
        """
        self.k_spec_dict[tag] = k_spec

    def pack(self):
        return json.dumps(self.k_spec_dict, sort_keys=True, cls=NpEncoder)

    def get_acronym(self):
        if self.acronym == "":
            for element in self.k_spec_dict.keys():
                self.acronym += self.engines[element].get_acronym()
        return self.acronym

    def bind(self):
        """
        binds the objects that actually compute the kernels
        these objects need to have .transform() method to compute 
        kernels from decriptor matrix [n_descriptors, n_samples]
        """
        # clear up the objects
        self.engines = {}
        for element in self.k_spec_dict.keys():
            self.engines[element] = self._call(self.k_spec_dict[element])
            self.k_spec_dict[element]['acronym'] = self.engines[element].get_acronym()

    def _call(self, k_spec):
        """
        call the specific kernel objects
        """
        if "type" not in k_spec.keys():
            raise ValueError("Did not specify the type of the kernel function.")
        if k_spec["type"] == "linear":
            return Kernel_Function_Linear(k_spec)
        if k_spec["type"] == "polynomial":
            return Kernel_Function_Polynomial(k_spec)
        if k_spec["type"] == "cosine":
            return Kernel_Function_Cosine(k_spec)
        else:
            raise NotImplementedError 

    def compute(self, desc_a, desc_b=None):
        """
        compute the global descriptor vector for a frame from atomic contributions
        Parameters
        ----------
        desc : array-like, shape=[n_descriptors, n_samples]
            design matrix

        Returns
        -------
        k_mat : array-like, shape=[n_samples, n_samples]
            design matrix
        """
        if desc_b is None:
            desc_b = desc_a

        n_a = len(desc_a)
        n_b = len(desc_b)

        k_mat = np.zeros((n_a,n_b), dtype=float)
        for element in self.k_spec_dict.keys():
            #print(np.shape(k_mat), np.shape(self.engines[element].transform(desc_a, desc_b)))
            k_mat += self.engines[element].transform(desc_a, desc_b)
            
        # this is not normalized!
        return k_mat

class Kernel_Function_Base:
    def __init__(self, k_spec):
        self.acronym = ""
        pass
    def get_acronym(self):
        # we use an acronym for each descriptor, so it's easy to find it and refer to it
        return self.acronym
    def transform(self, desc_a, desc_b):
        return []

class Kernel_Function_Linear(Kernel_Function_Base):
    def __init__(self, k_spec):
        self.acronym = "linear"
        try: 
            self.normalize =  k_spec['normalize']
        except:
            self.normalize = False
    def transform(self, desc_a, desc_b):
        if self.normalize and len(desc_a) == len(desc_b):
            return normalizekernel(np.dot(desc_a, desc_b.T))
        else:
            return np.dot(desc_a, desc_b.T)

class Kernel_Function_Polynomial(Kernel_Function_Base):
    def __init__(self, k_spec):
        self.acronym = "poly"
        self.d =  k_spec['d']
        try: 
            self.normalize =  k_spec['normalize']
        except:
            self.normalize = False
    def transform(self, desc_a, desc_b):
        if self.normalize and len(desc_a) == len(desc_b):
            return normalizekernel(np.power(np.dot(desc_a, desc_b.T),self.d))
        else:
            return np.power(np.dot(desc_a, desc_b.T),self.d)

class Kernel_Function_Cosine(Kernel_Function_Base):
    def __init__(self, k_spec):
        self.acronym = 'cos'
    def transform(self, desc_a, desc_b):
        from sklearn.metrics.pairwise import cosine_similarity as sk_cos_sim
        return sk_cos_sim(desc_a, desc_b)




"""
Tools for doing kernel PCA on environmental similarity
e.g.
kNN = np.genfromtxt(prefix+".k",skip_header=1)
proj = KernelPCA(kpca_d).fit_transform(kNN)

KernelPCA with precomputed kernel for now only!!!!!!!!!!!!
"""

import numpy as np
import scipy.linalg as salg


class KernelPCA:
    def __init__(self, n_components=2):
        """KernelPCA with precomputed kernel for now only

        Parameters
        ----------
        n_components : int, default=2
            number of dimensions to project to
        """

        # essential
        self.n_components = n_components

        # kernel and projection related properties
        self.vectors = None
        self.colmean = None
        self.mean = None
        self.center_kmat = None
        self._lambdas = None
        self._alphas = None
        self._m = None

        # state properties
        self._fitted = False

    @staticmethod
    def center_square(kernel):
        """Centering of a kernel matrix, with additional centering info

        Parameters
        ----------
        kernel : array-like, (shape=MxM)
            Kernel matrix

        Returns
        -------
        colmean : ndarray, (shape=M)
            mean of columns from kernel matrix
        mean : float
            mean of the entire kernel matrix
        centered_kernel : ndarray, (shape=MxM)
            kernel centered in feature space
        """
        colmean = np.mean(kernel, axis=1)
        mean = np.mean(colmean)
        centered_kernel = kernel - colmean - np.atleast_2d(colmean).T + mean
        return colmean, mean, centered_kernel

    def _center_test_kmat(self, ktest):
        """Centers a kernel matrix for projection

        Parameters
        ----------
        ktest : array_like, shape=(L,M)
            kernel matricx for the points to project to the lower dimensions
            Overwritten!!!

        Returns
        -------
        ktest :
            centered kernel matrix

        """
        test_colmean = np.atleast_2d(np.mean(ktest, axis=1)).T
        ktest += self.mean - self.colmean - test_colmean
        return ktest

    def fit(self, kmat):
        """Fit kernel PCA on the precomputed kernel matrix

        Notes
        -----
        - Keeps the kernel matrix intact
        - can be done only once on a given object, need to reinitialise if trying to run again

        Parameters
        ----------
        kmat : numpy.ndarray, shape=(M,M)
            kerenl matrix between the observations
            needs to be square and real, symmetric

        Returns
        -------
        """
        if self._fitted:
            raise RuntimeError('Kernel already fitted before, please reinitialise the object!')

        # check the shape and symmetry of the kernel matrix
        self._check_kmat(kmat, square=True)

        # save kmat and center
        self._m = len(kmat)
        self.colmean, self.mean, self.center_kmat = self.center_square(kmat)

        # calculation, ordering in descending order and scale by sqrt(lambda)
        self._lambdas, self._alphas = salg.eigh(self.center_kmat, eigvals=(self._m - self.n_components, self._m - 1))
        self._lambdas = np.flipud(self._lambdas)
        self._alphas = np.fliplr(self._alphas) / np.sqrt(self._lambdas)

        # the model is fitted then
        self._fitted = True

    def fit_transform(self, kmat, copy=True):
        """Fit kernel PCA on the precomputed kernel matrix & project to lower dimensions

        Notes
        -----
        - Keeps the kernel matrix intact
        - can be done only once on a given object, need to reinitialise if trying to run again

        Parameters
        ----------
        kmat : numpy.ndarray, shape=(M,M)
            kerenl matrix between the observations
            needs to be square and real, symmetric
        copy : bool, optional, default=True
            copy the kernel matrix or overwrite it, passed to self.transform()
            nb. the kernel matrix will be left centered if this is False

        Returns
        -------

        """
        self.fit(kmat)
        return self.transform(self.center_kmat, iscentered=True, copy=copy)

    def transform(self, ktest, iscentered=False, copy=True):
        """Transforms to the lower dimensions

        Parameters
        ----------
        ktest : array_like, shape=(L,M)
            kernel matrix of the test vectors with the training vectors
        iscentered : bool, optional, default=False
            if the kernel is centered already, mainly used for the fit_transform function
        copy : bool, optional, default=True
            copy the kernel matrix or overwrite it
            nb. the kernel matrix will be left centered if this is False

        Returns
        -------
        projected_vectors: numpy.array, shape=(L,n_components)
            projections of the given kernel wrt. the training kernel
        """
        if not self._fitted:
            raise RuntimeError("The model has not been fitted yet, please fit it and then use transform.")

        self._check_kmat(ktest, square=False)

        if iscentered:
            centered_ktest = ktest
        else:
            centered_ktest = self._center_test_kmat((ktest.copy() if copy else ktest))

        return np.dot(centered_ktest, self._alphas)

    def _check_kmat(self, kmat, square=True):
        """Check the kernel matrix

        checks if the matrix is real valued and its shape: square symmetric if square=True, LxM(self._m) otherwise

        Parameters
        ----------
        kmat : numpy.array
            kernel matrix to be checked
        square : bool, default=True
            if the matrix needs to be square and symmetric

        Returns
        -------

        """

        assert np.ndim(kmat) == 2
        assert np.isreal(kmat).all()

        if square:
            # shape (M,M)
            assert np.shape(kmat)[0] == np.shape(kmat)[1]
            # symmetric
            assert (kmat - kmat.T).max() < 1e-6
        else:
            # shape: (anything,M)
            assert np.shape(kmat)[1] == self._m

    def fit_vectors(self, vecs):
        """Fit Kernel PCA from vectors in the large dimension

        for future, not implemented yet
        """
        raise NotImplementedError

    def transform_vectors(self, vecs):
        """Fit Kernel PCA from vectors in the large dimension & project to lower dimension

        for future, not implemented yet
        """
        raise NotImplementedError



"""
Base class for regressions
"""

from sklearn.base import BaseEstimator, RegressorMixin
from .getscore import get_score


class RegressorBase(BaseEstimator, RegressorMixin):
    """
    Base class for regressions
    contains generic methods for computing errors
    """

    def __init__(self):
        """
        method specific
        """
        self._fitted = False

    def fit(self, X, y):
        """
        method specific
        """
        return self

    def predict(self, X, y):
        """
        method specific
        """
        pass

    def fit_predict(self, X, y, X_test):
        """
        Train the ridge regression model with the design matrix and trainLabel.
        Parameters
        ----------
        X : array-like, shape=[n_descriptors, n_samples]
        Input points.
        y : array-like, shape=[n_samples]
        Input points.
        X_test : array-like, shape=[n_descriptors, n_test_samples]
        Input points.
        """
        self.fit(X, y)
        return self.predict(X_test)

    def predict_error(self, X, y):
        """
        compute the score of the predictions compared with the true values
        Parameters
        ----------
        X : array-like, shape=[n_descriptors, n_samples]
        Input points.
        y : array-like, shape=[n_samples]
        Input points.
        """
        if not self._fitted:
            raise RuntimeError("The model has not been fitted yet, please fit it and then use predict.")

        # get the predictions for train set
        y_pred = self.predict(X)
        # compute the CV score
        y_error = get_score(y_pred, y)
        return y_pred, y_error

    def fit_predict_error(self, X, y, X_test, y_test):
        """
        Train the ridge regression model with the design matrix and trainLabel.
        and does prediction on test samples, and compute the error
        Parameters
        ----------
        X : array-like, shape=[n_descriptors, n_samples]
        Input points.
        y : array-like, shape=[n_samples]
        Input points.
        X_test : array-like, shape=[n_descriptors, n_test_samples]
        Input points.
        """
        self.fit(X, y)
        return self.predict_error(X_test, y_test)

    def get_train_test_error(self, X_train, y_train, X_test, y_test, verbose=True, return_pred=True):
        """
        train the model, and get the test error
        """
        fit_error = {}
        # compute the CV score for the train dataset
        y_pred, train_error = self.predict_error(X_train, y_train)
        if verbose: print("train score: ", train_error)
        fit_error['train_error'] = train_error

        # compute the CV score for the test dataset
        y_pred_test, test_error  = self.predict_error(X_test, y_test)
        if verbose: print("test score: ", test_error)
        fit_error['test_error'] = test_error

        if return_pred == True:
            return y_pred, y_pred_test, fit_error
        else:
            return fit_error

    def get_params(self, deep=True):
        pass

    def get_name(self):
        return type(self).__name__

    def set_params(self, deep=True):
        pass

    def pack(self):
        pass

    def unpack(self, state):
        pass

    def loads(self, state):
        pass





"""
Color functions
"""

import os
import numpy as np

from asaplib.data import ASAPXYZ


def set_color_function(fcolor='none', asapxyz=None, colorscol=0, n_samples=0, 
              peratom=False, project_atomic=False, use_atomic_species=None, color_from_zero=False, extensive=False):
    """ obtain the essential informations to define the colors of data points
    Parameters
    ----------
    fcolor: str
             the name of the file or the tag in the xyz to define the colors
    asapxyz: ASAPXYZ object, (optional)
    colorscol: int, (optional). 
              if the color file has more than one column, which column to use
    n_samples: int, (optional). 
              The number of data points
    peratom: bool
              return atomic color
    project_atomic: bool
              the samples are atomic descriptors
    use_atomic_species: int
              the atomic number of the selected species
    color_from_zero: bool
              set the min color to zero
    extensive: bool
              normalize the quatity by number of atoms
    """

    plotcolor = []
    plotcolor_atomic = []
    colorscale = [None, None]

    # if there is a file named "fcolor", we load it for the color scheme
    if os.path.isfile(fcolor):
        # load the column=colorscol for color functions
        try:
            loadcolor = np.genfromtxt(fcolor, dtype=float)
        except:
            raise IOError('Error in loading fcolor files for the color scheme')

        # print(np.shape(loadcolor))
        if colorscol > 0 or len(np.shape(loadcolor)) > 1:
            plotcolor = loadcolor[:, colorscol]
        else:
            plotcolor = loadcolor
        print('load file: ' + fcolor + ' for color schemes')

        if peratom or project_atomic:
            if asapxyz is None:
                raise IOError('Need the xyz so that we know the number of atoms in each frame')
            elif asapxyz.get_num_frames() == len(plotcolor):
                for index, natomnow in enumerate(asapxyz.get_natom_list_by_species(use_atomic_species)):
                    plotcolor_atomic = np.append(plotcolor_atomic, plotcolor[index] * np.ones(natomnow))
            elif asapxyz.get_total_natoms() == len(plotcolor):
                plotcolor_atomic = plotcolor
            else:
                raise ValueError('Length of the xyz trajectory is not the same as number of colors in the fcolor file')

    elif n_samples > 0 and (fcolor == None or fcolor == 'none' or fcolor == 'Index' or fcolor == 'index') and peratom == False:
        # we use the index as the color scheme
        plotcolor = np.arange(n_samples)
        fcolor = 'sample index'

    elif asapxyz is None:
        raise IOError('Cannot find the xyz or fcolor files for the color scheme')

    else:
        if fcolor == None or fcolor == 'none' or fcolor == 'Index' or fcolor == 'index':
            # we use the index as the color scheme
            plotcolor = np.arange(asapxyz.get_num_frames())
            fcolor = 'sample index'
            if peratom or project_atomic:
                for index, natomnow in enumerate(asapxyz.get_natom_list_by_species(use_atomic_species)):
                    plotcolor_atomic = np.append(plotcolor_atomic, plotcolor[index] * np.ones(natomnow))
        else:
            try:
                plotcolor = asapxyz.get_property(fcolor, extensive)
            except:
                raise ValueError('Cannot find the specified property from the xyz file for the color scheme')
            if peratom or project_atomic:
                try:
                    plotcolor_atomic = asapxyz.get_atomic_property(fcolor, extensive, [], use_atomic_species)
                    #print(np.shape(plotcolor_atomic))
                except:
                    raise ValueError('Cannot find the specified atomic property from the xyz file for the color scheme')

    if color_from_zero:
        # set the min to zero
        plotcolor -= np.ones(len(plotcolor))*np.nanmin(plotcolor)
        plotcolor_atomic -= np.ones(len(plotcolor_atomic))*np.nanmin(plotcolor)

    colorlabel = str(fcolor)
    if peratom and not project_atomic:
        # print(np.shape(plotcolor_atomic))
        colorscale = [np.nanmin(plotcolor_atomic), np.nanmax(plotcolor_atomic)]
        return plotcolor, np.asarray(plotcolor_atomic), colorlabel, colorscale
    elif project_atomic:
        colorscale = [None, None]
        return np.asarray(plotcolor_atomic), [], colorlabel, colorscale
    else:
        colorscale = [None, None]
        return plotcolor, [], colorlabel, colorscale


class COLOR_PALETTE:
    def __init__(self, style=1):
        if style == 1:
            self.pal = ["#FF34FF", "#FF4A46", "#008941", "#006FA6", "#A30059", "#0000A6", "#63FFAC", "#B79762",
                        "#004D43", "#8FB0FF", "#997D87", "#5A0007", "#809693",
                        "#1B4400", "#4FC601", "#3B5DFF", "#4A3B53", "#FF2F80", "#61615A", "#BA0900", "#6B7900",
                        "#00C2A0", "#FFAA92", "#FF90C9", "#B903AA", "#D16100", "#DDEFFF",
                        "#000035", "#7B4F4B", "#A1C299", "#300018", "#0AA6D8", "#013349", "#00846F", "#372101",
                        "#FFB500", "#C2FFED", "#A079BF", "#CC0744", "#C0B9B2", "#C2FF99",
                        "#001E09", "#00489C", "#6F0062", "#0CBD66", "#EEC3FF", "#456D75", "#B77B68", "#7A87A1",
                        "#788D66", "#885578", "#FAD09F", "#FF8A9A", "#D157A0", "#BEC459",
                        "#456648", "#0086ED", "#886F4C", "#34362D", "#B4A8BD", "#00A6AA", "#452C2C", "#636375",
                        "#A3C8C9", "#FF913F", "#938A81", "#575329", "#00FECF", "#B05B6F",
                        "#8CD0FF", "#3B9700", "#04F757", "#C8A1A1", "#1E6E00", "#7900D7", "#A77500", "#6367A9",
                        "#A05837", "#6B002C", "#772600", "#D790FF", "#9B9700", "#549E79",
                        "#FFF69F", "#201625", "#72418F", "#BC23FF", "#99ADC0", "#3A2465", "#922329", "#5B4534",
                        "#FDE8DC", "#404E55", "#0089A3", "#CB7E98", "#A4E804", "#324E72", "#6A3A4C", "#7A4900"]
        elif style == 2:
            self.pal = ["#30a2da", "#fc4f30", "#e5ae38", "#6d904f", "#8b8b8b", "#006FA6", "#A30059", "#af8dc3",
                        "#922329", "#1E6E00"]

        self.n_color = len(self.pal)

    def __getitem__(self, arg):  # color cycler
        assert arg > -1, "???"
        return self.pal[arg % self.n_color]



"""
Functions for performing Fast Fourier transforms
"""

import numpy as np

"""
# Fourier analysis of x(t) to obtain $x(\omega)$.
## $FT(x(t)) = x(\omega)$
## $S_x(\omega)=FT(<x(0)x(t)>) = x(\omega) \cdot x(\omega) $

#  In reality we have a list $\{ x_{1...n}\}$
## $A_k = \sum_{m=0}^{n-1} x_m \exp(-2\pi i \dfrac{m k }{n})$
## $x(\omega = \dfrac{2\pi}{n\Delta t}k) = \Delta t A_k$ 

# Fourier analysis of v(t) to obtain $v(\omega)$.
## $FFT(v(t)) = v(\omega)=i\omega x(\omega)$
## $FFT(<x(0)v(t)>) = -x(\omega) \cdot v(\omega) $
##  $FFT(<x(0)v(t)>) = i \omega x(\omega) \cdot x(\omega) $
## $FFT(<v(0)v(t)>) = v(\omega) \cdot v(\omega) $
##  $FFT(<v(0)v(t)>) = -(i \omega)^2 x(\omega) \cdot x(\omega) = \omega^2 x(\omega) \cdot x(\omega)$

EXAMPLES:
testfx = fftranform(txv[:,[0,1]],len(txv)) fftranform(txv[:,[0,1]],len(txv)) 
testreversefx = ifftranform(testfx) 

testftcxx = rfftcrosscorr(txv[:,[0,1]],txv[:,[0,1]],10000)
"""


def smooth(y, box_pts):
    box = np.ones(box_pts) / box_pts
    y_smooth = np.convolve(y, box, mode='same')
    return y_smooth


def fftranform(x, dlen=10000):
    # input: a list of real signals {x} with timestep dt
    # output: x(omega)

    # make sure that we have odd number of signals as it makes fft easier
    if dlen % 2 == 0:
        dlen -= 1
        # x.append(x[0])
    # this is the FFT expansion coeficients assuming inputs are real numbers
    # https://docs.scipy.org/doc/numpy-1.13.0/reference/routines.fft.html#module-numpy.fft
    dt = x[1, 0] - x[0, 0]  # assume the timestep is constant
    window = len(x) // dlen
    # print dlen
    omega0 = 2.0 * np.pi / (dlen - 1) / dt
    xomega = np.zeros((dlen, 2), dtype=np.complex_)
    xomega[0:dlen // 2 + 1, 0] = np.arange(dlen // 2 + 1) * omega0
    xomega[dlen // 2 + 1:, 0] = np.arange(dlen // 2, 0, -1) * omega0 * -1

    for i in range(window):
        dx = x[i * dlen:(i + 1) * dlen, 1]
        Ax = np.fft.fft(dx[:], axis=0)
        # print len(Ax)
        xomega[:, 1] += Ax * dt
    xomega[:, 1] /= window
    # return [omega, A(omega)]
    """ when numpy does fft, A[1:n/2] contains the positive-frequency terms, 
    and A[n/2+1:] contains the negative-frequency terms, 
    in order of decreasingly negative frequency.
    """
    return xomega


def ifftranform(xomega):
    # first retrieve the descrete rFFT coefficients
    # note that numpy uses a normalization factor of 1/n here but not during the forward fft
    omega0 = xomega[1, 0] - xomega[0, 0]
    dt = np.pi / (len(xomega) // 2) / omega0
    Ax = xomega[:, 1] / dt
    x = np.fft.ifft(Ax[:], axis=0)
    return np.column_stack((dt * np.arange(len(x)), x))


def fftcrosscorr(x, y, dlen=10000):
    # make sure that we have odd number of signals as it makes fft easier
    if dlen % 2 == 0:
        dlen -= 1
        # x.append(x[0])
        # y.append(y[0])

    # the fft coecofficents of the crosscorrelation function c_xy(t)
    dt = x[1, 0] - x[0, 0]  # assume the timestep is constant
    window = len(x) // dlen
    omega0 = 2.0 * np.pi / (dlen - 1) / dt
    cxyomega = np.zeros((dlen, 2), dtype=np.complex_)
    cxyomega[0:dlen // 2 + 1, 0] = np.arange(dlen // 2 + 1) * omega0
    cxyomega[dlen // 2 + 1:, 0] = np.arange(dlen // 2, 0, -1) * omega0 * -1

    for i in range(window):
        dx = x[i * dlen:(i + 1) * dlen, 1]
        dy = y[i * dlen:(i + 1) * dlen, 1]
        Ax = np.fft.fft(dx[:], axis=0)
        Ay = np.fft.fft(dy[:], axis=0)

        cxyomega[:, 1] += np.conjugate(Ax[:]) * Ay[:] / dlen * dt

    for i in range(window - 1):
        dx = x[i * dlen + dlen // 2:(i + 1) * dlen + dlen // 2, 1]
        dy = y[i * dlen + dlen // 2:(i + 1) * dlen + dlen // 2, 1]
        Ax = np.fft.fft(dx[:], axis=0)
        Ay = np.fft.fft(dy[:], axis=0)

        cxyomega[:, 1] += np.conjugate(Ax[:]) * Ay[:] / dlen * dt

    cxyomega[:, 1] /= (window * 2 - 1)
    return cxyomega


def rfftranform(x, dlen=10000):
    # input: a list of real signals {x} with timestep dt
    # output: x(omega)

    # this is the FFT expansion coeficients assuming inputs are real numbers
    # https://docs.scipy.org/doc/numpy-1.13.0/reference/routines.fft.html#module-numpy.fft
    dt = x[1, 0] - x[0, 0]  # assume the timestep is constant
    window = len(x) // dlen
    # print dlen
    omega0 = 2.0 * np.pi / dlen / dt
    xomega = np.zeros((dlen // 2 + 1, 2), dtype=np.complex_)
    xomega[:, 0] = np.arange(dlen // 2 + 1) * omega0

    for i in range(window):
        dx = x[i * dlen:(i + 1) * dlen, 1]
        Ax = np.fft.rfft(dx[:], axis=0)
        # print len(Ax)
        xomega[:, 1] += Ax * dt
    xomega[:, 1] /= window
    # return [omega, A(omega)]
    """ when numpy does fft, A[1:n/2] contains the positive-frequency terms, 
    and A[n/2+1:] contains the negative-frequency terms, 
    in order of decreasingly negative frequency.
    Hoever, When the DFT is computed for purely real input, 
    the output is Hermitian-symmetric, 
    i.e. the negative frequency terms are just the complex conjugates of the corresponding positive-frequency terms, 
    and the negative-frequency terms are therefore redundant. 
    This function does not compute the negative frequency terms, 
    and the length of the transformed axis of the output is therefore n//2 + 1
    """
    return xomega


def irfftranform(xomega):
    # first retrieve the descrete rFFT coefficients
    # note that numpy uses a normalization factor of 1/n here but not during the forward fft
    omega0 = xomega[1, 0] - xomega[0, 0]
    dt = np.pi / len(xomega) / omega0
    Ax = xomega[:, 1] / dt
    x = np.fft.irfft(Ax[:], axis=0)
    return np.column_stack((dt * np.arange(len(x)), x))


def rfftcrosscorr(x, y, dlen=10000):
    # the fft coecofficents of the crosscorrelation function c_xy(t)
    dt = x[1, 0] - x[0, 0]  # assume the timestep is constant
    window = len(x) // dlen
    omega0 = 2.0 * np.pi / dlen / dt
    cxyomega = np.zeros((dlen // 2 + 1, 2), dtype=np.complex_)
    cxyomega[:, 0] = np.arange(dlen // 2 + 1) * omega0

    for i in range(window):
        dx = x[i * dlen:(i + 1) * dlen, 1]
        dy = y[i * dlen:(i + 1) * dlen, 1]
        Ax = np.fft.rfft(dx[:], axis=0)
        Ay = np.fft.rfft(dy[:], axis=0)

        cxyomega[:, 1] += np.conjugate(Ax[:]) * Ay[:] / dlen * dt

    for i in range(window - 1):
        dx = x[i * dlen + dlen // 2:(i + 1) * dlen + dlen // 2, 1]
        dy = y[i * dlen + dlen // 2:(i + 1) * dlen + dlen // 2, 1]
        Ax = np.fft.rfft(dx[:], axis=0)
        Ay = np.fft.rfft(dy[:], axis=0)

        cxyomega[:, 1] += np.conjugate(Ax[:]) * Ay[:] / dlen * dt

    cxyomega[:, 1] /= (window * 2 - 1)
    return cxyomega



"""
sparsifier class
"""

from asaplib.compressor import random_split, fps, CUR_deterministic

class Sparsifier:
    def __init__(self, sparse_mode):
        """
        Object handing the sparsification of data
        Parameters
        ----------
        sparse_mode: str
                     Type of the sparse mode
        """
        self._possible_modes = ['fps', 'cur', 'random', 'sequential']
        if sparse_mode.lower() not in self._possible_modes:
            raise NotImplementedError("Do not recognize the selected sparsification mode. \
                                    Use ([cur], [fps], [random],[sequential]).")
        else:
            self.sparse_mode = sparse_mode.lower()

    def _check(self, n_sparse, n_total):
        # sanity check
        if n_sparse > n_total:
            print("the number of representative structure is too large, please select n <= ", n_total)

    def sparsify(self, desc_or_ntotal, n_or_ratio, sparse_param=0):
        """
        Function handing the sparsification of data
        Parameters
        ----------
        desc_or_ntotal: np.matrix or int
                        Either a design matrix [n_sample, n_desc],
                        or simply the total number of samples
        n_or_ratio: int or float 
                  Either the number or the fraction of sparsified points
        sparse_param: int
                additional parameter that may be needed for the specific sparsifier used

        Returns
        ----------
        sbs: list
        a list of the indexes for the sparsified points
        """
        if isinstance(desc_or_ntotal, int):
            n_total = desc_or_ntotal
            input_desc = False
        else:
            desc = desc_or_ntotal
            n_total = len(desc_or_ntotal)
            input_desc = True

        if n_or_ratio == 1 or isinstance(n_or_ratio, float):
            n_sparse = n_total * n_or_ratio
        elif isinstance(n_or_ratio, int):
            n_sparse = n_or_ratio
        else:
            raise ValueError("the sparsification ratio/number should be a float or int.")
 
        self._check(n_sparse, n_total)

        if self.sparse_mode == 'fps':
            if not input_desc: 
                raise ValueError("fps needs design matrix")
            sbs, _ = fps(desc, n_sparse, int(sparse_param))
        elif self.sparse_mode == 'cur':
            if not input_desc:
                raise ValueError("cur needs design matrix")
            import numpy as np
            cov = np.dot(np.asmatrix(desc), np.asmatrix(desc).T)
            sbs, _ = CUR_deterministic(cov, n_sparse)
        elif self.sparse_mode == 'random':
            _, sbs = random_split(n_total, n_sparse/n_total)
        elif self.sparse_mode == 'sequential':
            sbs = range(n_sparse)
        else:
            raise ValueError("sparse mode not right")

        return sbs



import numpy as np

from .base import RegressorBase

class RidgeRegression(RegressorBase):
    _pairwise = True

    def __init__(self, jitter):
        # Weights of the rr model
        self.alpha = None
        self.jitter = jitter  # noise level^2
        self.coninv = None  # inverse of the covariance matrix
        self._fitted = False

    def fit(self, X, y):
        """
        Train the ridge regression model with the design matrix and trainLabel.
        Parameters
        ----------
        X : array-like, shape=[n_descriptors, n_samples]
        Input points.
        y : array-like, shape=[n_samples]
        Input points.
        """
        #print("a total of ", np.shape(X), "column")

        # calculate covariance matrix
        COV = np.dot(X.T, X)
        #print("computing covariance matrix with shape:", np.shape(COV))

        reg = np.eye(COV.shape[0]) * self.jitter
        # self.coninv = np.linalg.inv(COV+reg)
        # self.alpha = np.dot(self.coninv, np.dot(X.T,y))
        self.alpha = np.linalg.solve(COV + reg, np.dot(X.T, y))
        self._fitted = True

    def predict(self, X):
        '''desc.shape is expected as [n_descriptors, n_samples]'''
        if not self._fitted:
            raise RuntimeError("The model has not been fitted yet, please fit it and then use predict.")
        return np.dot(X, self.alpha.flatten()).reshape((-1))

    def get_params(self, deep=True):
        return dict(sigma=self.jitter)

    def set_params(self, params, deep=True):
        self.jitter = params['jitter']
        self.alpha = None

    def pack(self):
        state = dict(weights=self.alpha, jitter=self.jitter)
        return state

    def unpack(self, state):
        self.alpha = state['weights']
        err_m = 'jitter are not consistent {} != {}'.format(self.jitter, state['jitter'])
        assert self.jitter == state['jitter'], err_m

    def loads(self, state):
        self.alpha = state['weights']
        self.jitter = state['jitter']



"""
Farthest Point Sampling methods for sparsification
"""

import numpy as np


def fps(x, d=0, r=None):
    """
    Farthest Point Sampling

    Parameters
    ----------
    x: np.matrix
       [n_samples, n_dim] coordinates of all the samples to be sparsified.
    d: int
       number of samples to keep
    r: int
       starting from sample of index r

    Returns
    -------
    sample_index: list
        a list of selected samples, remaining error

    """

    if d == 0:
        d = len(x)
    n = len(x)
    iy = np.zeros(d, int)
    # faster evaluation of Euclidean distance
    n2 = np.einsum("ai,ai->a", x, x)
    if r is None:
        iy[0] = np.random.randint(0, n)
    else:
        iy[0] = r
    dl = n2 + n2[iy[0]] - 2 * np.dot(x, x[iy[0]])
    lmin = np.zeros(d)
    for i in range(1, d):
        iy[i] = np.argmax(dl)
        lmin[i - 1] = dl[iy[i]]
        nd = n2 + n2[iy[i]] - 2 * np.dot(x, x[iy[i]])
        dl = np.minimum(dl, nd)
    return iy, lmin


def fast_fps(x, d=0, r=None):
    if d == 0:
        d = len(x)
    n = len(x)
    iy = np.zeros(d, int)
    # faster evaluation of Euclidean distance
    n2 = np.einsum("ai,ai->a", x, x)
    if r is None:
        iy[0] = np.random.randint(0, n)
    else:
        iy[0] = r

    ldmin = n2 + n2[iy[0]] - 2 * np.dot(x, x[iy[0]])
    ivor = np.zeros(n, int)  # voronoi assignment of each point
    rvor = np.zeros(d)  # maximum distance within a voronoi ("voronoi radius")
    lmin = np.zeros(d)

    yx = np.zeros((d, len(x[0])))
    yx[0] = x[iy[0]]
    rvor[0] = np.max(ldmin)
    ndist = 0
    for i in tqdm_notebook(range(1, d)):
        inew = np.argmax(ldmin)  # index of the maximum minimum distance
        iy[i] = inew

        dnew = ldmin[inew]  # keep track of the maximum minimum distance
        lmin[i - 1] = dnew

        yx[i] = x[inew]

        # now gets the distance between this point and all the points selected
        # up to now (could be eliminated by caching!)
        syd = np.sqrt(np.abs(n2[iy[:i]] + n2[inew] -
                             2 * np.dot(yx[:i], x[inew])))
        sdnew = np.sqrt(dnew)

        # list of active voronoi (we check which of the voronoi cells contain points
        # that might contain points closer than dmax)
        iactive = np.where(rvor[:i] > dnew + syd[:i]
                           * (syd[:i] - 2 * sdnew))[0]

        rvor[iactive] = 0
        for j in xrange(n):
            if ivor[j] in iactive:
                if ldmin[j] > dnew + syd[ivor[j]] * (syd[ivor[j]] - 2 * sdnew):
                    dj = n2[j] + n2[inew] - 2 * np.dot(x[j], x[inew])
                    ndist += 1
                    if dj < ldmin[j]:
                        ivor[j] = i
                        ldmin[j] = dj
                    rvor[ivor[j]] = max(rvor[ivor[j]], ldmin[j])
    print(" computed ", ndist, "distances")
    return iy, lmin, ivor, rvor



"""
tools for generating hyperparameters for ACSF descriptors
"""
import json
import numpy as np

from .univeral_length_scales import uni_length_scales, system_pair_bond_lengths, round_sigfigs
from ..io import NpEncoder

"""
Automatically generate the hyperparameters of ACSF descriptors for arbitrary elements and combinations.

## Heuristics:
  * Get the length scales of the system from
    * maximum bond length (from equilibrium bond length in lowest energy 2D or 3D structure)
    * minimal bond length (from shortest bond length of any equilibrium structure, including dimer)
  * Apply a scaling for these length scales
    * acsf cutoff = maximum bond length * 1.56 * scalerange
    * rmin = minimal bond length, distance in Angstrom to the first nearest neighbor.
         Eliminates the symmetry functions that investigate the space between 0 and rmin.
    * N = the number of each type of SFs. Determined by
          N = min(int(sharpness*(cutoff-rmin)/0.5), 5)
## Example
"""

def universal_acsf_hyper(global_species, facsf_param, dump=True, verbose=True):

    """
    format:
    acsf_js = {'acsf1': {'type': 'ACSF',
                        'cutoff': 2.0,
                        'g2_params': [[1, 1], [1, 2], [1, 3]],
                        'g4_params': [[1, 1, 1], [1, 2, 1], [1, 1, -1], [1, 2, -1]]}}
    """
    
    if facsf_param == 'smart' or facsf_param == 'Smart' or facsf_param == 'SMART':
        acsf_js = gen_default_acsf_hyperparameters(list(global_species), scalerange=1.2, sharpness=1.0)
    elif facsf_param == 'minimal' or facsf_param == 'Minimal' or facsf_param == 'MINIMAL':
        acsf_js = gen_default_acsf_hyperparameters(list(global_species), scalerange=0.85, sharpness=1.0)
    elif facsf_param == 'longrange' or facsf_param == 'Longrange' or facsf_param == 'LONGRANGE':
        acsf_js = gen_default_acsf_hyperparameters(list(global_species),scalerange=1.8, sharpness=1.2)
    elif isinstance(facsf_param, float):
        acsf_js = gen_default_acsf_hyperparameters(list(global_species),scalerange=1.0, sharpness=1.0, cutoff=facsf_param)
    else:
        raise IOError('Did not specify acsf parameters. You can use [smart/minimal/longrange] or set an explicit cutoff.')
        
    if verbose:  
        print(acsf_js)
    if dump:
        with open('smart-acsf-parameters', 'w') as jd:
            json.dump(acsf_js, jd, cls=NpEncoder)
    return acsf_js

def gen_default_acsf_hyperparameters(Zs, scalerange=1.0, sharpness=1.0, verbose=False, cutoff=None):
    """
    Parameters
    ----------
    Zs : array-like, list of atomic species
    scalerange: type=float, scale the cutoffs of the SFs.
    sharpness: type=float, sharpness factor for atom_width, default=1.0,
                larger sharpness means more resolution, and more SFs will be generated.
    verbose: type=bool, default=False, more descriptions of what has been done.
    """

    # check if the element is in the look up table 
    for Z in Zs:
        if str(Z) not in uni_length_scales:
            raise RuntimeError("key Z {} not present in length_scales table".format(Z))

    shortest_bond, longest_bond = system_pair_bond_lengths(Zs, uni_length_scales)
    if verbose:
        print(Zs, "range of bond lengths", shortest_bond, longest_bond)
        
    # cutoffs & shortest length
    if cutoff is None:
        cutoff = max(float(round_sigfigs(longest_bond * 1.3 * float(scalerange), 2)),2.0)
    rmin = shortest_bond
    N = int(sharpness*(cutoff-rmin)/0.5)
    if verbose:
        print("Considering cutoff and rmin", cutoff, rmin)

    index = np.arange(N+1, dtype=float)
    shift_array = cutoff*(1./N)**(index/(len(index)-1))
    eta_array = 1./shift_array**2.

    _2_body_params = []
    for eta in eta_array:
        # G2 with no shift
        if 3*np.sqrt(1/eta) > rmin:
            _2_body_params.append([float(round_sigfigs(eta, 2)), 0.])
            if verbose:
                for fel in Zs:
                    for sel in Zs: 
                        print("symfunction_short %s 2 %s %.4f 0.000 %.3f" %(fel, sel, eta, cutoff))
    for i in range(len(shift_array)-1):
        # G2 with shift
        eta = 1./((shift_array[N-i] - shift_array[N-i-1])**2)
        if shift_array[N-i] + 3*np.sqrt(1/eta) > rmin:
            _2_body_params.append([float(round_sigfigs(eta,2)), float(round_sigfigs(shift_array[N-i],2))])
            if verbose: 
                for fel in Zs:
                    for sel in Zs:
                        print("symfunction_short %s 2 %s %.4f %.3f %.3f" %(fel, sel, eta, shift_array[N-i], cutoff))

    eta_array = np.logspace(-3,0,N//2)
    zeta_array = [1.000, 4.000, 16.000]
    
    _3_body_params = []
    for eta in eta_array:
        for zeta in zeta_array:
            if 3*np.sqrt(1/eta) > rmin:
                _3_body_params.append([float(round_sigfigs(eta,2)), float(round_sigfigs(zeta,2)), 1])
                _3_body_params.append([float(round_sigfigs(eta,2)), float(round_sigfigs(zeta,2)), -1])

    if verbose:
        for fel in Zs:
            ang_Zs = list(Zs)
            for sel in Zs:
                for tel in ang_Zs:
                    print("# symfunctions for type %s 3 %s %s" %(fel, sel, tel))
                    for eta in eta_array:
                        for zeta in zeta_array:
                            if 3*np.sqrt(1/eta) > rmin:
                                print("symfunction_short %s 3 %s %s %.4f  1.000 %.3f %.3f" %(fel, sel, tel, eta, zeta, cutoff))
                                print("symfunction_short %s 3 %s %s %.4f -1.000 %.3f %.3f" %(fel, sel, tel, eta, zeta, cutoff))
                            
    acsf_js = { 'acsf-'+str(cutoff):{'type': 'ACSF',
    'cutoff': cutoff,
    'g2_params': _2_body_params,
    'g4_params': _3_body_params }}

    return acsf_js



"""
tools for doing PCA
e.g.
desc = np.genfromtxt(prefix+".desc")
proj = PCA(pca_d).fit_transform(desc)
"""

import numpy as np
import scipy.linalg as salg
from sklearn.preprocessing import StandardScaler


class PCA:
    def __init__(self, n_components=2, scalecenter=True):
        """PCA with precomputed descriptor matrix

        Parameters
        ----------
        n_components : int, default=2
            number of dimensions to project to
        """

        # essential
        self.n_components = n_components
        # options for scale and center the descriptor matrix
        self.scalecenter = scalecenter
        self.scaler = None

        # eigenvectors for projections
        self.pvec = None

        # state properties
        self._fitted = False

    def scalecenter_matrix(self, desc):
        """Scaling and Centering of a design matrix, with additional centering info

        Parameters
        ----------
        desc : array-like, shape=[n_descriptors, n_samples]
            design matrix

        Returns
        -------
        """
        self.scaler = StandardScaler()
        print(self.scaler.fit(desc))
        return self.scaler.transform(desc)

    def centering(self, desc):
        # calculate the mean of each column
        M_desc = np.mean(desc.T, axis=1)
        # center columns by subtracting column means
        C_desc = desc - M_desc
        return C_desc

    def fit(self, desc):
        """Fit PCA on the precomputed descriptor matrix

        Parameters
        ----------
        desc: array-like, shape=[n_descriptors, n_samples]
        design matrix

        Returns
        -------
        """
        if self._fitted:
            raise RuntimeError('PCA already fitted before, please reinitialise the object!')

        print("Start PCA for a design matrix with shape ", np.shape(desc))

        # scale and center
        if self.scalecenter:
            C_desc = self.scalecenter_matrix(desc)
        else:
            # center columns by subtracting column means
            C_desc = self.centering(desc)

        # calculate covariance matrix of centered matrix
        COV = np.cov(C_desc.T)
        print("computing covariance matrix with shape:", np.shape(COV))

        print("  And now we build a projection ")
        eval, evec = salg.eigh(COV, eigvals=(len(COV) - self.n_components, len(COV) - 1))
        eval = np.flipud(eval)
        evec = np.fliplr(evec)

        self.pvec = evec.copy()
        for i in range(self.n_components):
            self.pvec[:, i] *= 1. / np.sqrt(eval[i])
        print("Done, super quick. ")

        # the model is fitted then
        self._fitted = True

    def transform(self, desc_test):
        """Transforms to the lower dimensions

        Parameters
        ----------
        desc_test : array_like, shape=[n_descriptors, n_samples]
            design matrix of the new samples

        Returns
        -------
        projected_vectors: numpy.array, shape=(L,n_components)
            projections of the design matrix on low dimension
        """
        if not self._fitted:
            raise RuntimeError("The model has not been fitted yet, please fit it and then use transform.")

        if self.scalecenter:
            desc_test = self.scaler.transform(desc_test)

        return np.dot(desc_test, self.pvec)

    def fit_transform(self, desc):
        """Fit PCA on the design matrix & project to lower dimensions

        Parameters
        ----------
        desc_test : array_like, shape=[n_descriptors, n_samples]
            design matrix of the new samples

        Returns
        -------
        projected_vectors: numpy.array, shape=(L,n_components)
            projections of the design matrix on low dimension

        """
        self.fit(desc)
        return self.transform(desc)



"""
Class for storing and handling design matrices
"""

import numpy as np
import json
from yaml import dump as ydump

from ..io import randomString,  NpEncoder
from ..compressor import random_split,exponential_split, LCSplit, ShuffleSplit
from ..compressor import Sparsifier
from ..fit import LC_SCOREBOARD

class Design_Matrix:
    """extended design matrix class

    Parameters
    ----------
    X : array-like, shape=[n_samples,n_desc]
    Input points.
    y : array-like, shape=[n_samples]
    label for every point
    testratio: float, ratio of the test fraction
    z : array-like, shape=[n_samples]
    additional label for every point
    tags: array-like, strings, shape=[n_samples]
    additional tags for each data point
    """
    def __init__(self, X=[], y=[], whiten=True, test_ratio=0, random_seed=42, z=[], tags=[]):
        # sanity checks
        if len(y) > 0 and len(X) != len(y):
            raise ValueError('Length of the labels y is not the same as the design matrix X')
        
        if whiten:
            X = self._whiten(X)

        self.n_sample = len(X)
        self.test_ratio = test_ratio
        self.random_seed = random_seed

        # We do a train/test split already here, to ensure the subsequent analysis is prestine
        self.train_list, self.test_list = random_split(self.n_sample, test_ratio, random_seed)
        self.X_train = X[self.train_list]
        self.X_test = X[self.test_list]
        self.n_train = len(self.X_train)
        self.n_test = len(self.X_test)

        if len(y) > 0:
            self.y_train = y[self.train_list]
            self.y_test = y[self.test_list]
        if len(z) > 0:
            self.z_train = z[self.train_list]
            self.z_test = z[self.test_list]
        if len(tags) > 0:
            self.tags_train = tags[self.train_list]
            self.tags_test = tags[self.test_list]

        # this stores the index of representative data
        self.sbs = range(self.n_train)
        self.n_sparse = len(self.X_train) # no sparsification for now

        # these store the results of the fit using external learners
        self.fit_error_by_learner = {} # fitting errors
        self.lc_by_learner = {} # learning curves

    def _whiten(self, X, scale = True, addbias = True):
        """ scale & center """
        if scale:
            from sklearn.preprocessing import StandardScaler
            scaler = StandardScaler()
            X = scaler.fit_transform(X)  # normalizing the features
        # add bias
        if addbias:
            X_bias = np.ones((np.shape(X)[0], np.shape(X)[1] + 1))
            X_bias[:, 1:] = X
            return X_bias
        else:
            return X

    def save_state(self, filename, mode='yaml'):
        """output json or yaml file"""
        if mode == 'yaml':
            with open(filename+'-fit-errors.yaml', 'w') as yd:
                ydump(self.fit_error_by_learner, yd, sort_keys=True)
            with open(filename+'-lc.yaml', 'w') as yd:
                ydump(self.lc_by_learner, yd, sort_keys=True)
        else:
            with open(filename+'-fit-errors.json', 'w') as jd:
                json.dump(self.fit_error_by_learner, jd, sort_keys=True, cls=NpEncoder)
            with open(filename+'-lc.json', 'w') as jd:
                json.dump(self.lc_by_learner, jd, sort_keys=True, cls=NpEncoder)

    def sparsify(self, n_sparse=None, sparse_mode='fps'):
        """
        select representative data points using the design matrix

        Parameters
        ----------
        n_sparse: int
                  number of representative points
                  n_sparse == None means 5% of the data
                  n_sparse < 0 means no sparsification
        sparse_mode: str
                  Methods to use for sparsification [cur], [fps], [random]
        """
        
        # set default value of n_sparse
        if n_sparse is None:
            n_sparse = max(10, self.n_train // 20)

        # sparsification
        if n_sparse > 0:
            sparsifier = Sparsifier(sparse_mode)
            self.sbs = sparsifier.sparsify(desc, n_sparse)

    def get_sparsified_matrix(self):
        if len(self.y_train) > 0:
            return self.X_train[self.sbs], self.y_train[self.sbs]
        else:
            return self.X_train[self.sbs], []

    def compute_fit(self, learner, tag=None, store_results=True, plot=True):
        """
        Fit the design matrix X and the values y using a learner

        Parameters
        ----------
        learner: a learner object
                 e.g. RidgeRegression
                 needs to have .fit(), .predict(), .get_train_test_error(), .fit_predict_error() methods
        tag: str
              The name of this learner
        """
        # sanity checks
        if len(self.y_train) < 1 or len(self.X_train) != len(self.y_train):
            raise ValueError('Length of the labels y is not the same as the design matrix X')

        if tag is None: tag = randomString(6)

        # fit the model
        learner.fit(self.X_train[self.sbs], self.y_train[self.sbs])

        if store_results:
            y_pred, y_pred_test, fit_error = learner.get_train_test_error(self.X_train[self.sbs], self.y_train[self.sbs], self.X_test, self.y_test, verbose=True, return_pred=True)
            self.fit_error_by_learner[tag] = {"error": fit_error}

        if plot:
            from matplotlib import pyplot as plt
            fig, ax = plt.subplots()
            ax.plot(self.y_train[self.sbs], y_pred, 'b.', label='train')
            ax.plot(self.y_test, y_pred_test, 'r.', label='test')
            ax.legend()
            ax.set_title('Fits using: '+str(tag))
            ax.set_xlabel('actual y')
            ax.set_ylabel('predicted y')
            return fig, ax

    def compute_learning_curve(self, learner, tag=None, lc_points=8, lc_repeats=8, randomseed=42, verbose=True):
        """
        Fit the learning curve using a learner

        Parameters
        ----------
        lc_points: int
                   the number of points on the learning curve
        lc_repeats: int
                the number of sub-samples to take when compute the learning curve
        learner: a learner object
                 e.g. RidgeRegression
                 needs to have .fit(), .predict(), .get_train_test_error(), .fit_predict_error() methods
        tag: str
                The name of this learner
        """
        if tag is None: tag = randomString(6)

        if lc_points < 1:
            print("The number of points on the learning curve < 1. Skip computing learning curve")
            return
        if lc_repeats < 1:
            print("The number sub-samples to take when compute the learning curve < 1. Skip computing learning curve")
            return

        # set lower and upper bound of the LC
        train_sizes = exponential_split(max(20,self.n_train//1000), len(self.sbs), lc_points)
        print("Learning curves using train sizes: ", train_sizes)
        lc_stats = lc_repeats * np.ones(lc_points-1, dtype=int)
        lc = LCSplit(ShuffleSplit, n_repeats=lc_stats, train_sizes=train_sizes[:-1], test_size=None, random_state=randomseed)

        lc_scores = LC_SCOREBOARD(train_sizes)
        for lctrain, _ in lc.split(self.X_train[self.sbs]):
            Ntrain = len(lctrain)
            lc_X_train = self.X_train[self.sbs][lctrain, :]
            lc_y_train = self.y_train[self.sbs][lctrain]
            # here we always use the same test set
            _, lc_score_now = learner.fit_predict_error(lc_X_train, lc_y_train, self.X_test, self.y_test)
            lc_scores.add_score(Ntrain, lc_score_now)
        # for the whole data set
        _, lc_score_now = learner.fit_predict_error(self.X_train[self.sbs], self.y_train[self.sbs], self.X_test, self.y_test)
        lc_scores.add_score(self.n_train, lc_score_now)

        self.lc_by_learner.update({tag:lc_scores.fetch_all()})
        if verbose: print("LC results: ", {tag:lc_scores.fetch_all()})
        return lc_scores





"""
Functions for IO and type conversions
"""

import argparse
import json
import click
import ast

import numpy as np
import random
import string

def extract_from_nested_dict(key, var):
    if hasattr(var,'iteritems'):
        for k, v in var.iteritems():
            if k == key:
                yield v
            if isinstance(v, dict):
                for result in extract_from_nested_dict(key, v):
                    yield result
            elif isinstance(v, list):
                for d in v:
                    for result in extract_from_nested_dict(key, d):
                        yield result

def str2bool(v):
    """

    Parameters
    ----------
    v

    Returns
    -------

    """
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def randomString(stringLength=8):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(stringLength))

def list2str(input_list):
    output_str = ""
    for l in input_list:
        output_str += str(l)
        output_str += "-"
    return output_str

class ConvertStrToList(click.Option):
    def type_cast_value(self, ctx, value):
        try:
            value = str(value)
            assert value.count('[') == 1 and value.count(']') == 1
            list_as_str = value.replace('"', "'").split('[')[1].split(']')[0]
            list_of_items = [item.strip().strip("'") for item in list_as_str.split(',')]
            return list_of_items
        except Exception:
            raise click.BadParameter(value)

class PythonLiteralOption(click.Option):

    def type_cast_value(self, ctx, value):
        try:
            return ast.literal_eval(value)
        except:
            raise click.BadParameter(value)

class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)

class NpDecoder(json.JSONDecoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpDecoder, self).default(obj)
